from astropy.coordinates import SkyCoord
from astropy import units as u
import pytest

from . import assert_widget_image, wait_for_test
from ..conftest import RUNNING_ON_CI


WAIT_TIME = 5 if RUNNING_ON_CI else 1
M42 = SkyCoord.from_name('M42')


def check_silent_output(capsys):
    out, err = capsys.readouterr()
    assert out.strip() == ""
    assert err.strip() == ""


class TestWWTWidget:

    def test_settings(self, capsys, wwt_qt_client):
        wwt_qt_client.constellation_figures = True
        wwt_qt_client.constellation_figures = False
        wait_for_test(wwt_qt_client, WAIT_TIME)
        check_silent_output(capsys)

    def test_methods(self, capsys, wwt_qt_client):
        wwt_qt_client.center_on_coordinates(M42, fov=10 * u.deg)
        wait_for_test(wwt_qt_client, WAIT_TIME)
        check_silent_output(capsys)

    def test_coordinates(self, capsys, wwt_qt_client):
        wwt_qt_client.center_on_coordinates(M42, fov=10 * u.deg)
        wait_for_test(wwt_qt_client, WAIT_TIME)
        assert M42.separation(wwt_qt_client.get_center()).arcsec < 1.e-6
        wait_for_test(wwt_qt_client, WAIT_TIME)
        check_silent_output(capsys)

    def test_annotations(self, capsys, wwt_qt_client):
        circle = wwt_qt_client.add_circle()
        circle.opacity = 0.8
        circle.set_center(M42)
        wait_for_test(wwt_qt_client, WAIT_TIME)
        check_silent_output(capsys)


def test_full(tmpdir, wwt_qt_client_isolated):
    """
    Test a whole Qt session, with image comparison along the way. We don't
    immediately fail if the images disagree, to allow building up a full
    database of disagreeing images to help understand what's going on.
    """

    failures = []
    wwt = wwt_qt_client_isolated

    # Step 0

    wwt.foreground_opacity = 1.
    gc = SkyCoord(0, 0, unit=('deg', 'deg'), frame='icrs')
    wwt.center_on_coordinates(gc, 60 * u.deg)
    wwt.crosshairs = False

    wait_for_test(wwt, WAIT_TIME, for_render=True)

    msg = assert_widget_image(tmpdir, wwt, 'qt_full_step0.png', fail_now=False)
    if msg:
        failures.append(msg)

    # Step 1

    gc = SkyCoord(0, 0, unit=('deg', 'deg'), frame='galactic')
    wwt.center_on_coordinates(gc, 60 * u.deg)

    wait_for_test(wwt, WAIT_TIME, for_render=True)

    msg = assert_widget_image(tmpdir, wwt, 'qt_full_step1.png', fail_now=False)
    if msg:
        failures.append(msg)

    # Step 2

    wwt.constellation_boundary_color = 'red'
    wwt.constellation_figure_color = 'green'
    wwt.constellation_selection_color = 'blue'
    wwt.constellation_boundaries = True
    wwt.constellation_figures = True

    wait_for_test(wwt, WAIT_TIME, for_render=True)

    msg = assert_widget_image(tmpdir, wwt, 'qt_full_step2.png', fail_now=False)
    if msg:
        failures.append(msg)

    # Step 3

    wwt.constellation_selection = True
    wwt.ecliptic = True
    wwt.grid = True

    wait_for_test(wwt, WAIT_TIME, for_render=True)

    msg = assert_widget_image(tmpdir, wwt, 'qt_full_step3.png', fail_now=False)
    if msg:
        failures.append(msg)

    # Step 4

    wwt.foreground = 'SFD Dust Map (Infrared)'

    wait_for_test(wwt, WAIT_TIME, for_render=True)

    msg = assert_widget_image(tmpdir, wwt, 'qt_full_step4.png', fail_now=False)
    if msg:
        failures.append(msg)

    # Step 5

    wwt.foreground = "Black Sky Background"
    wwt.background = "Black Sky Background"
    wwt.foreground_opacity = 0

    wwt.center_on_coordinates(gc, 30 * u.deg)

    coord = SkyCoord(5, 0.5, unit=('deg', 'deg'), frame='galactic')

    circle1 = wwt.add_circle()
    circle1.set_center(coord)
    circle1.radius = 10 * u.pixel
    circle1.line_width = 5 * u.pixel
    circle1.line_color = 'green'
    circle1.fill = False
    circle1.opacity = 0.5

    coord = SkyCoord(-5, -0.5, unit=('deg', 'deg'), frame='galactic')

    circle2 = wwt.add_circle()
    circle2.set_center(coord)
    circle2.radius = 2 * u.degree
    circle2.line_width = 5 * u.pixel
    circle2.line_color = 'green'
    circle2.fill = True
    circle2.fill_color = 'orange'
    circle2.opacity = 1

    coord = SkyCoord([1, 4, 0], [0, 0, -5], unit=('deg', 'deg'), frame='galactic')

    poly = wwt.add_polygon()
    poly.add_point(coord[0])
    poly.add_point(coord[1])
    poly.add_point(coord[2])
    poly.fill = True
    poly.line_color = 'red'
    poly.fill_color = 'yellow'
    poly.line_width = 2 * u.pixel

    coord = SkyCoord([10, 5, 2], [5, 2, 2], unit=('deg', 'deg'), frame='galactic')

    polyline = wwt.add_line()
    polyline.add_point(coord[0])
    polyline.add_point(coord[1])
    polyline.add_point(coord[2])
    polyline.color = 'green'
    polyline.width = 3 * u.pixel

    wait_for_test(wwt, WAIT_TIME, for_render=True)

    msg = assert_widget_image(tmpdir, wwt, 'qt_full_step5.png', fail_now=False)
    if msg:
        failures.append(msg)

    # Summarize

    if failures:
        pytest.fail('\n'.join(failures), pytrace=False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
    "numpydoc",
]

templates_path = ["_templates"]

source_suffix = ".rst"

master_doc = "index"

project = "pywwt"
author = "AAS WorldWide Telescope team, John ZuHone"
copyright = "2017-2021, " + author

# The short X.Y version.
from pywwt import __version__ as version

# The full version, including alpha/beta/rc tags.
release = version

language = None

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

pygments_style = "sphinx"

todo_include_todos = False

html_theme = "bootstrap-astropy"
html_theme_options = {
    "logotext1": "pywwt",
    "logotext2": "",
    "logotext3": ":docs",
    "astropy_project_menubar": False,
}
html_static_path = ["_static"]
html_sidebars = { '**': ['localtoc.html', 'searchbox.html'] }

htmlhelp_basename = "pywwtdoc"

intersphinx_mapping = {
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "matplotlib": (
        "https://matplotlib.org/",
        (None, "https://matplotlib.org/objects.inv"),
    ),
    "numpy": (
        "https://docs.scipy.org/doc/numpy/",
        (None, "https://numpy.org/doc/stable/objects.inv"),
    ),
    "python": (
        "https://docs.python.org/3/",
        (None, "http://data.astropy.org/intersphinx/python3.inv"),
    ),
    "scipy": (
        "https://docs.scipy.org/doc/scipy/reference/",
        (None, "http://data.astropy.org/intersphinx/scipy.inv"),
    ),
    "toasty": (
        "https://toasty.readthedocs.io/en/latest/",
        (None, "https://toasty.readthedocs.io/en/latest/objects.inv"),
    ),
    "traitlets": (
        "https://traitlets.readthedocs.io/en/latest/",
        (None, "https://traitlets.readthedocs.io/en/latest/objects.inv"),
    ),
}

numpydoc_show_class_members = False

nitpicky = True
nitpick_ignore = [("py:class", "ipywidgets.widgets.domwidget.DOMWidget")]

default_role = "obj"

html_logo = "images/logo.png"

linkcheck_retries = 5
linkcheck_timeout = 10
linkcheck_ignore = ["https://en.wikipedia.org/wiki/Aircraft_principal_axes"]



import numpy as np
import pytz
from astropy.io import fits
from astropy.coordinates import ICRS
from astropy.time import Time
from datetime import datetime
from reproject import reproject_interp
from reproject.mosaicking import find_optimal_celestial_wcs

__all__ = ["sanitize_image"]


def sanitize_image(image, output_file, overwrite=False, hdu_index=None, **kwargs):
    """
    Transform a FITS image so that it is in equatorial coordinates with a TAN
    projection and floating-point values, all of which are required to work
    correctly in WWT at the moment.

    Image can be a filename, an HDU, or a tuple of (array, WCS).
    """

    # In case of a FITS file with more than one HDU, we need to choose one
    if isinstance(image, str):
        with fits.open(image) as hdul:
            if hdu_index is not None:
                image = hdul[hdu_index]
            else:
                for hdu in hdul:
                    if (
                        hasattr(hdu, "shape")
                        and len(hdu.shape) > 1
                        and type(hdu) is not fits.hdu.table.BinTableHDU
                    ):
                        break
                image = hdu
            transform_to_wwt_supported_fits(image, output_file, overwrite)
    else:
        transform_to_wwt_supported_fits(image, output_file, overwrite)


def transform_to_wwt_supported_fits(image, output_file, overwrite):
    # Workaround because `reproject` currently only accepts 2D inputs. This is a
    # hack and it would be better to update reproject to do this processing.
    # Also, this logic is copy/pasting `toasty.collection.SimpleFitsCollection`.

    import warnings
    from reproject.utils import parse_input_data

    with warnings.catch_warnings():
        # Sorry, Astropy, no one cares if you fixed the FITS.
        warnings.simplefilter("ignore")
        data, wcs = parse_input_data(image)

    if wcs.naxis != 2:
        if not wcs.has_celestial:
            raise Exception(
                f"cannot process input `{image}`: WCS cannot be reduced to 2D celestial"
            )

        full_wcs = wcs
        wcs = full_wcs.celestial

        # note: get_axis_types returns axes in FITS order, innermost first
        keep_axes = [
            t.get("coordinate_type") == "celestial"
            for t in full_wcs.get_axis_types()[::-1]
        ]

        for axnum, (keep, axlen) in enumerate(zip(keep_axes, data.shape)):
            if not keep and axlen != 1:
                # This is a non-celestial axis that we need to drop, but its
                # size is not one. So in principle the user should tell us which
                # plane to chose. We can't do that here, so just complain --
                # that's better than giving a hard error since this way the user
                # can at least see *something*.
                warnings.warn(
                    f"taking first plane (out of {axlen}) in non-celestial image axis #{axnum} in input `{image}`"
                )

        data = data[tuple(slice(None) if k else 0 for k in keep_axes)]

    image = (data, wcs)

    # End workaround.

    wcs, shape_out = find_optimal_celestial_wcs([image], frame=ICRS(), projection="TAN")
    array = reproject_interp(image, wcs, shape_out=shape_out, return_footprint=False)
    fits.writeto(
        output_file, array.astype(np.float32), wcs.to_header(), overwrite=overwrite
    )


def validate_traits(cls, traits):
    """
    Helper function to ensure user-provided trait names match those of the
    class they're being used to instantiate.
    """
    mismatch = [key for key in traits if key not in cls.trait_names()]
    if mismatch:
        raise KeyError(
            "Key{0} {1} do{2}n't match any layer trait name".format(
                "s" if len(mismatch) > 1 else "",
                mismatch,
                "" if len(mismatch) > 1 else "es",
            )
        )


def ensure_utc(tm, str_allowed):
    """
    Helper function to convert a time object (Time, datetime, or UTC string
    if str_allowed == True) into UTC before passing it to WWT.
    str_allowed is True for wwt.set_current_time (core.py) and False for TableLayer's 'time_att' implementation (layers.py).
    """

    if tm is None:
        utc_tm = datetime.utcnow().astimezone(pytz.UTC).isoformat()

    elif isinstance(tm, datetime):
        if tm.tzinfo is None:
            utc_tm = pytz.utc.localize(tm).isoformat()
        elif tm.tzinfo == pytz.UTC:
            utc_tm = tm.isoformat()
        else:  # has a non-UTC time zone
            utc_tm = tm.astimezone(pytz.UTC).isoformat()

    elif isinstance(tm, Time):
        utc_tm = tm.to_datetime(pytz.UTC).isoformat()

    else:
        if str_allowed:  # is an ISOT string
            dt = Time(tm, format="isot").to_datetime(pytz.UTC)
            utc_tm = dt.isoformat()
        else:
            raise ValueError("Time must be a datetime or astropy.Time object")

    return utc_tm



import sys
import uuid
import tempfile
from os import path
import shutil
import asyncio

import nest_asyncio
from pathlib import Path

import re

if sys.version_info[0] == 2:  # noqa
    from io import BytesIO as StringIO
else:
    from io import StringIO

import warnings
from base64 import b64encode
from collections import OrderedDict

import numpy as np
from astropy.io import fits
from matplotlib.pyplot import cm
from matplotlib.colors import Colormap
from astropy import units as u
import astropy.units.imperial  # noqa: F401
from astropy.table import Column
from astropy.table import Table
from astropy.time import Time
from datetime import datetime
import toasty

from traitlets import HasTraits, validate, observe
from .traits import Color, Bool, Float, Unicode, AstropyQuantity, Any, to_hex
from .utils import sanitize_image, validate_traits, ensure_utc

__all__ = [
    "CatalogHipsLayer",
    "ImageLayer",
    "LayerManager",
    "TableLayer",
]

VALID_FRAMES = [
    "sky",
    "ecliptic",
    "galactic",
    "sun",
    "mercury",
    "venus",
    "earth",
    "mars",
    "jupiter",
    "saturn",
    "uranus",
    "neptune",
    "pluto",
    "moon",
    "io",
    "europa",
    "ganymede",
    "callisto",
]

VALID_LON_UNITS = {u.deg: "degrees", u.hour: "hours", u.hourangle: "hours"}

# NOTE: for cartesian coordinates, we can also allow custom units
VALID_ALT_UNITS = {
    u.m: "meters",
    u.imperial.foot: "feet",
    u.imperial.inch: "inches",
    u.imperial.mile: "miles",
    u.km: "kilometers",
    u.au: "astronomicalUnits",
    u.lyr: "lightYears",
    u.pc: "parsecs",
    u.Mpc: "megaParsecs",
}

VALID_ALT_TYPES = ["depth", "altitude", "distance", "seaLevel", "terrain"]

VALID_MARKER_TYPES = ["gaussian", "point", "circle", "square", "pushpin"]

VALID_MARKER_SCALES = ["screen", "world"]

VALID_STRETCHES = ["linear", "log", "power", "sqrt", "histeq"]

VALID_COLORMAPS = [
    "viridis",
    "plasma",
    "inferno",
    "magma",
    "cividis",
    "Greys",
    "gray",
    "Purples",
    "Blues",
    "Greens",
    "Oranges",
    "Reds",
    "RdYlBu",
]

UI_COLORMAPS = OrderedDict(
    [
        ("Viridis", "viridis"),
        ("Plasma", "plasma"),
        ("Black to white", "gray"),
        ("White to black", "Greys"),
        ("Inferno", "inferno"),
        ("Magma", "magma"),
        ("Cividis", "cividis"),
        ("Red-Yellow-Blue", "RdYlBu"),
        ("Purples", "Purples"),
        ("Blues", "Blues"),
        ("Greens", "Greens"),
        ("Oranges", "Oranges"),
        ("Reds", "Reds"),
    ]
)

# Save string types for validating ISOT strings in time series tables
if sys.version_info[0] == 2:
    STR_TYPE = basestring  # noqa
    NP_STR_TYPE = np.string_
else:
    STR_TYPE = str
    NP_STR_TYPE = np.unicode_

# The following are columns that we add dynamically and internally, so we need
# to make sure they have unique names that won't clash with existing columns
SIZE_COLUMN_NAME = str(uuid.uuid4())
CMAP_COLUMN_NAME = str(uuid.uuid4())
TIME_COLUMN_NAME = str(uuid.uuid4())


def guess_lon_lat_columns(colnames):
    """
    Given column names in a table, return the columns to use for lon/lat, or
    None/None if no high confidence possibilities.
    """

    # Do all the checks in lowercase
    colnames_lower = [colname.lower() for colname in colnames]

    for lon, lat in [("ra", "dec"), ("lon", "lat"), ("lng", "lat")]:

        # Check first for exact matches

        lon_match = [colname == lon for colname in colnames_lower]
        lat_match = [colname == lat for colname in colnames_lower]

        if sum(lon_match) == 1 and sum(lat_match) == 1:
            return colnames[lon_match.index(True)], colnames[lat_match.index(True)]

        # Next check for columns that start with specified names

        lon_match = [colname.startswith(lon) for colname in colnames_lower]
        lat_match = [colname.startswith(lat) for colname in colnames_lower]

        if sum(lon_match) == 1 and sum(lat_match) == 1:
            return colnames[lon_match.index(True)], colnames[lat_match.index(True)]

        # We don't check for cases where lon/lat are inside the name but not at
        # the start since that might be e.g. for proper motions (pm_ra) or
        # errors (dlat).

    return None, None


def guess_xyz_columns(colnames):
    """
    Given column names in a table, return the columns to use for x/y/z, or
    None/None/None if no high confidence possibilities.
    """

    # Do all the checks in lowercase
    colnames_lower = [colname.lower() for colname in colnames]

    for x, y, z in [("x", "y", "z")]:

        # Check first for exact matches

        x_match = [colname == x for colname in colnames_lower]
        y_match = [colname == y for colname in colnames_lower]
        z_match = [colname == z for colname in colnames_lower]

        if sum(x_match) == 1 and sum(y_match) == 1 and sum(z_match) == 1:
            return (
                colnames[x_match.index(True)],
                colnames[y_match.index(True)],
                colnames[z_match.index(True)],
            )

        # Next check for columns that start with specified names

        x_match = [colname.startswith(x) for colname in colnames_lower]
        y_match = [colname.startswith(y) for colname in colnames_lower]
        z_match = [colname.startswith(z) for colname in colnames_lower]

        if sum(x_match) == 1 and sum(y_match) == 1 and sum(z_match) == 1:
            return (
                colnames[x_match.index(True)],
                colnames[y_match.index(True)],
                colnames[z_match.index(True)],
            )

    return None, None, None


def pick_unit_if_available(unit, valid_units):
    # Check for equality rather than just identity
    for valid_unit in valid_units:
        if unit == valid_unit:
            return valid_unit
    return unit


def csv_table_win_newline(table):
    """
    Helper function to get Astropy tables as ASCII CSV with Windows line
    endings
    """
    s = StringIO()
    table.write(s, format="ascii.basic", delimiter=",", comment=False)
    s.seek(0)
    # Replace single \r or \n characters with \r\n
    return re.sub(r"(?<![\r\n])(\r|\n)(?![\r\n])", "\r\n", s.read())


class LayerManager(object):
    """
    A simple container for layers.
    """

    def __init__(self, parent=None):
        self._layers = []
        self._parent = parent

    def add_image_layer(
        self,
        image=None,
        hdu_index=None,
        verbose=True,
        name=None,
        force_hipsgen=False,
        force_tan=False,
        **kwargs
    ):
        """
        Add an image layer to the current view. This method can display any
        image data or FITS or list of FITS from the local environment. In case
        of a large FITS (> 20 MB), the file is preprocessed into tiles using
        the `toasty` library for smooth & fast visualization. You can also use
        this method to combine multiple FITS files into a single image set.

        Parameters
        ----------
        image : str or list of str or :class:`~astropy.io.fits.ImageHDU` or tuple
            The image to show, which should be given either as a filename,
            a list of FITS file paths, an :class:`~astropy.io.fits.ImageHDU`
            object, or a tuple of the form ``(array, wcs)`` where ``array`` is
            a Numpy array and ``wcs`` is an astropy :class:`~astropy.wcs.WCS`
            object
        hdu_index : optional int or list of int, defaults to None
            Use this parameter to specify which HDU to tile. If the *image*
            input is a list of FITS, you can specify the hdu_index of each FITS
            by using a list of integers like this: [0, 2, 1]. If hdu_index is
            not set, toasty will use the first HDU with tilable content in each
            FITS.
        verbose : optional boolean, defaults True
            If true, progress messages will be printed as the FITS files are
            being processed.
        name : optional str, defaults to None
            Use this parameter to set a display name for the image set.
            If not set, pywwt will set a name based on the file name of the
            provided image.
        force_hipsgen : optional boolean, defaults to False
            Force the input FITS to be tiled with HiPSgen. If this and
            *force_tan* are set to False, this method will figure out when to
            split the input into tiles and also which type of projection to
            use. Tangential projection for smaller angular areas and HiPSgen
            larger regions of the sky.
        force_tan : optional boolean, defaults to False
            Force the input FITS to be tiled with a tangential projection. If
            this and *force_hipsgen* are set to False, this method will figure
            out when to split the input into tiles and also which type of
            projection to use. Tangential projection for smaller angular areas
            and HiPSgen larger regions of the sky.
        kwargs
            Additional keyword arguments can be used to set properties on the
            image layer or settings for the `toasty` tiling process. Common
            toasty settings include ``out_dir``, ``override``, and ``blankval``.
            See `toasty.tile_fits`.

        Returns
        -------
        layer : :class:`~pywwt.layers.ImageLayer` or a subclass thereof
        """

        if (
            isinstance(image, astropy.io.fits.ImageHDU)
            or isinstance(image, astropy.io.fits.PrimaryHDU)
            or isinstance(image, astropy.io.fits.HDUList)
        ):
            unused_file_name = self._get_unused_fits_name()
            image.writeto(unused_file_name)
            image = unused_file_name
        if isinstance(image, str):
            image = [image]
        if isinstance(image, list):
            force_untiled = False
            if "force_untiled" in kwargs:
                force_untiled = kwargs["force_untiled"]
                kwargs.pop("force_untiled", None)
            if not force_untiled and (
                force_hipsgen
                or force_tan
                or len(image) > 1
                or Path(image[0]).stat().st_size > 20e6
            ):  # 20 MB
                nest_asyncio.apply()
                loop = asyncio.get_event_loop()
                return loop.run_until_complete(
                    self._tile_and_serve(
                        fits_list=image,
                        hdu_index=hdu_index,
                        cli_progress=verbose,
                        display_name=name,
                        force_hipsgen=force_hipsgen,
                        force_tan=force_tan,
                        **kwargs
                    )
                )
            else:
                return self._create_and_add_image_layer(
                    image=image[0], hdu_index=hdu_index, name=name, **kwargs
                )

        return self._create_and_add_image_layer(
            image=image, hdu_index=hdu_index, name=name, **kwargs
        )

    def _create_and_add_image_layer(self, image, **kwargs):
        kwargs = self._remove_toasty_keywords(**kwargs)
        layer = self._parent._create_image_layer(image=image, **kwargs)
        self._add_layer(layer)
        return layer

    def add_preloaded_image_layer(self, url=None, **kwargs):
        """
        Add an image layer to the current view, based on its data URL.

        Parameters
        ----------
        url : str
            The URL of the image, as registered with the WWT frontend.
        kwargs
            Additional keyword arguments can be used to set properties on the
            image layer.

        Returns
        -------
        layer : :class:`~pywwt.layers.ImageLayer` or a subclass thereof
        """
        layer = self._parent._create_image_layer(url=url, **kwargs)
        self._add_layer(layer)
        return layer

    # TODO this is not future proof
    def _remove_toasty_keywords(self, **kwargs):
        kwargs.pop("blankval", None)
        kwargs.pop("override", None)
        kwargs.pop("out_dir", None)
        return kwargs

    def _get_unused_fits_name(self):
        file_index = 1
        while Path("hdu_fits_{}".format(file_index)).is_file():
            file_index += 1
        return "hdu_fits_{}".format(file_index)

    async def _tile_and_serve(
        self,
        fits_list,
        hdu_index=None,
        cli_progress=True,
        display_name=None,
        force_hipsgen=False,
        force_tan=False,
        **kwargs
    ):
        with warnings.catch_warnings():
            # Avoid annoying AstroPy FITS-fixed warnings
            warnings.simplefilter("ignore")
            out_dir, builder = toasty.tile_fits(
                fits_list,
                hdu_index=hdu_index,
                cli_progress=cli_progress,
                force_hipsgen=force_hipsgen,
                force_tan=force_tan,
                **kwargs
            )

        kwargs = self._remove_toasty_keywords(**kwargs)
        url = self._parent._serve_tree(path=out_dir)

        self._parent.load_image_collection(url=url + "index.wtml", remote_only=True)

        if display_name is None:
            display_name = builder.imgset.name

        image_layer = self.add_preloaded_image_layer(
            url + builder.imgset.url, name=display_name, **kwargs
        )
        if builder.imgset.pixel_cut_low > 0 or builder.imgset.pixel_cut_low < 0:
            image_layer.vmin = builder.imgset.pixel_cut_low
            image_layer.vmax = builder.imgset.pixel_cut_high
            image_layer._data_min = builder.imgset.data_min
            image_layer._data_max = builder.imgset.data_max

        return image_layer

    def add_table_layer(self, table=None, frame="Sky", **kwargs):
        """
        Add a data layer to the current view

        Parameters
        ----------
        table : :class:`~astropy.table.Table`
            The table containing the data to show.
        frame : str
            The reference frame to use for the data. This should be either
            ``'Sky'``, ``'Ecliptic'``, ``Galactic``, or the name of a planet
            or a natural satellite in the solar system.
        kwargs
            Additional keyword arguments can be used to set properties on the
            table layer.

        Returns
        -------
        layer : :class:`~pywwt.layers.TableLayer`
        """

        # Validate frame
        if frame.lower() not in VALID_FRAMES:
            raise ValueError(
                "frame should be one of {0}".format(
                    "/".join(sorted(str(x) for x in VALID_FRAMES))
                )
            )
        frame = frame.capitalize()

        if table is not None:
            layer = TableLayer(self._parent, table=table, frame=frame, **kwargs)
        else:
            # NOTE: in future we may allow different arguments such as e.g.
            # orbit=, hence why we haven't made this a positional argument.
            raise ValueError("The table argument is required")
        self._add_layer(layer)
        return layer

    async def add_hips_catalog_layer(self, name, **kwargs):
        """
        Add a HiPS catalog layer to the current view.

        Parameters
        ----------
        name : str
            Name of the HiPS catalog to display. You can list them using the widget's
            :attr:`~pywwt.BaseWWTWidget.available_hips_catalog_names` attribute
        **kwargs
            Additional keyword arguments can be used to set properties on the
            catalog HiPS layer.

        Returns
        -------
        layer : :class:`~pywwt.layers.CatalogHipsLayer`

        Notes
        -----
        This function is asynchronous because the engine needs to download the
        information about the catalog and return its metadata to the Python code.
        """

        name = self._validate_hips_catalog_name(name)
        model_id = str(uuid.uuid4())

        fut = self._parent._send_into_future(
            event="layer_hipscat_load",
            name=name,
            tableId=model_id,
        )
        reply = await fut

        layer = CatalogHipsLayer(self._parent, model_id, reply)
        self._add_layer(layer)

        with layer.hold_trait_notifications():
            for key, value in kwargs.items():
                if not layer.has_trait(key):
                    raise TypeError("unexpected keyword argument '{}'".format(key))
                setattr(layer, key, value)

        return layer

    def _validate_hips_catalog_name(self, name):
        if name not in self._parent.available_hips_catalog_names:
            raise ValueError(
                "name should be one of {0}".format(
                    " /// ".join(
                        sorted(
                            str(x) for x in self._parent.available_hips_catalog_names
                        )
                    )
                )
            )
        return name

    def add_data_layer(self, *args, **kwargs):
        """
        Deprecated, use ``add_table_layer`` instead.
        """
        warnings.warn(
            "add_data_layer has been deprecated, use " "add_table_layer instead",
            DeprecationWarning,
        )
        return self.add_table_layer(*args, **kwargs)

    def _add_layer(self, layer):
        if layer in self._layers:
            raise ValueError("layer already exists in layer manager")
        self._layers.append(layer)
        layer._manager = self

    def remove_layer(self, layer):
        """
        Remove a layer from the view.

        Parameters
        ----------
        layer : :class:`TableLayer` or :class:`ImageLayer`
            The layer to remove.
        """

        if layer not in self._layers:
            raise ValueError("layer not in layer manager")
        layer.remove()
        # By this point, the call to remove() above may already have resulted
        # in the layer getting removed, so we check first if it's still present.
        if layer in self._layers:
            self._layers.remove(layer)

    def __len__(self):
        return len(self._layers)

    def __iter__(self):
        for layer in self._layers:
            yield layer

    def __getitem__(self, item):
        return self._layers[item]

    def __str__(self):
        if len(self) == 0:
            return "Layer manager with no layers"
        else:
            s = "Layer manager with {0} layers:\n\n".format(len(self))
            for ilayer, layer in enumerate(self._layers):
                s += "  [{0}]: {1}\n".format(ilayer, layer)
            return s

    __repr__ = __str__

    def _serialize_state(self):
        layer_states = []
        for layer in self._layers:
            layer_states.append(layer._serialize_state())

        return layer_states

    def _save_all_data_for_serialization(self, dir):
        for layer in self._layers:
            layer._save_data_for_serialization(dir)


class TableLayer(HasTraits):
    """
    A layer where the data is stored in an :class:`~astropy.table.Table`
    """

    coord_type = Unicode(
        "spherical",
        help="Whether to give the coordinates "
        "in spherical or rectangular coordinates",
    ).tag(wwt="coordinatesType")

    # Attributes for spherical coordinates

    lon_att = Unicode(help="The column to use for the longitude " "(`str`)").tag(
        wwt="lngColumn"
    )
    lon_unit = Any(
        help="The units to use for longitude " "(:class:`~astropy.units.Unit`)"
    ).tag(wwt="raUnits")
    lat_att = Unicode(help="The column to use for the latitude " "(`str`)").tag(
        wwt="latColumn"
    )

    alt_att = Unicode(help="The column to use for the altitude " "(`str`)").tag(
        wwt="altColumn"
    )
    alt_unit = Any(
        help="The units to use for the altitude " "(:class:`~astropy.units.Unit`)"
    ).tag(wwt="altUnit")
    alt_type = Unicode(help="The type of altitude (`str`)").tag(wwt="altType")

    # Attributes for cartesian coordinates

    x_att = Unicode(help="The column to use for the x " "coordinate").tag(
        wwt="xAxisColumn"
    )
    y_att = Unicode(help="The column to use for the y " "coordinate").tag(
        wwt="yAxisColumn"
    )
    z_att = Unicode(help="The column to use for the z " "coordinate").tag(
        wwt="zAxisColumn"
    )
    xyz_unit = Any(help="The units to use for the x/y/z " "positions").tag(
        wwt="cartesianScale"
    )

    # NOTE: we deliberately don't link size_att to sizeColumn because we need
    # to compute the sizes ourselves based on the min/max and then use the
    # resulting column.
    size_att = Unicode(help="The column to use for the marker size " "(`str`)").tag(
        wwt=None
    )
    size_vmin = Float(
        None,
        help="The minimum point size. "
        "Found automagically once size_att is set "
        "(`float`)",
        allow_none=True,
    ).tag(wwt=None)
    size_vmax = Float(
        None,
        help="The maximum point size. "
        "Found automagically once size_att is set "
        "(`float`)",
        allow_none=True,
    ).tag(wwt=None)

    cmap_att = Unicode(help="The column to use for the colormap " "(`str`)").tag(
        wwt=None
    )
    cmap_vmin = Float(
        None,
        help="The minimum level of the colormap. Found "
        "automagically once cmap_att is set (`float`)",
        allow_none=True,
    ).tag(wwt=None)
    cmap_vmax = Float(
        None,
        help="The maximum level of the colormap. Found "
        "automagically once cmap_att is set (`float`)",
        allow_none=True,
    ).tag(wwt=None)
    cmap = Any(
        cm.viridis,
        help="The Matplotlib colormap " "(:class:`matplotlib.colors.ListedColormap`)",
    ).tag(wwt=None)

    # Visual attributes

    size_scale = Float(
        10, help="The factor by which to scale the size " "of the points (`float`)"
    ).tag(wwt="scaleFactor")
    color = Color("white", help="The color of the markers " "(`str` or `tuple`)").tag(
        wwt="color"
    )
    opacity = Float(1, help="The opacity of the markers " "(`str`)").tag(wwt="opacity")

    marker_type = Unicode("gaussian", help="The type of marker " "(`str`)").tag(
        wwt="plotType"
    )
    marker_scale = Unicode(
        "screen",
        help="Whether the scale is " "defined in world or pixel coordinates " "(`str`)",
    ).tag(wwt="markerScale")

    far_side_visible = Bool(
        False,
        help="Whether markers on the far side "
        "of a 3D object are visible "
        "(`bool`)",
    ).tag(wwt="showFarSide")

    # NOTE: we deliberately don't link time_att to startDateColumn here
    # because we need to compute a new times column based on time_att before
    # passing the result on to WWT
    time_att = Unicode(help="The column to use for time (`str`)").tag(wwt=None)
    time_series = Bool(
        False, help="Whether the layer contains time series " "elements (`bool`)"
    ).tag(wwt="timeSeries")
    time_decay = AstropyQuantity(
        16 * u.day,
        help="How long a time series "
        "point takes to fade away after appearing (0 "
        "if never) "
        "(:class:`~astropy.units.Quantity`)",
    ).tag(wwt="decay")

    # TODO: support:
    # xAxisColumn
    # yAxisColumn
    # zAxisColumn
    # xAxisReverse
    # yAxisReverse
    # zAxisReverse

    def __init__(
        self,
        parent=None,
        table=None,
        frame=None,
        table_from_wwt_engine=False,
        id=None,
        **kwargs
    ):
        self.table = table
        self.notify_changes = True

        # Validate frame
        if frame.lower() not in VALID_FRAMES:
            raise ValueError(
                "frame should be one of {0}".format(
                    "/".join(sorted(str(x) for x in VALID_FRAMES))
                )
            )
        self.frame = frame.capitalize()

        self.parent = parent

        if id is None:
            id = str(uuid.uuid4())
        self.id = id

        # Attribute to keep track of the manager, so that we can notify the
        # manager if a layer is removed.
        self._manager = None
        self._removed = False

        if not table_from_wwt_engine:
            self._initialize_layer()

            # Force defaults
            self._on_trait_change({"name": "alt_type", "new": self.alt_type})
            self._on_trait_change({"name": "size_scale", "new": self.size_scale})
            self._on_trait_change({"name": "color", "new": self.color})
            self._on_trait_change({"name": "opacity", "new": self.opacity})
            self._on_trait_change({"name": "marker_type", "new": self.marker_type})
            self._on_trait_change({"name": "marker_scale", "new": self.marker_scale})
            self._on_trait_change(
                {"name": "far_side_visible", "new": self.far_side_visible}
            )
            self._on_trait_change({"name": "size_att", "new": self.size_att})
            self._on_trait_change({"name": "cmap_att", "new": self.cmap_att})
            self._on_trait_change({"name": "time_att", "new": self.time_att})
            self._on_trait_change({"name": "time_series", "new": self.time_series})
            self._on_trait_change({"name": "time_decay", "new": self.time_decay})

            self._on_trait_change({"name": "cmap", "new": self.cmap})

        self.observe(self._on_trait_change, type="change")

        # Check that all kwargs are valid -- throws error if not
        validate_traits(self, kwargs)

        super(TableLayer, self).__init__(**kwargs)

        if not table_from_wwt_engine:
            if kwargs.get("coord_type") == "rectangular":

                x_guess, y_guess, z_guess = guess_xyz_columns(self.table.colnames)

                if "x_att" not in kwargs:
                    self.x_att = x_guess or self.table.colnames[0]

                if "y_att" not in kwargs:
                    self.y_att = y_guess or self.table.colnames[1]

                if "z_att" not in kwargs:
                    self.z_att = z_guess or self.table.colnames[2]

            else:

                lon_guess, lat_guess = guess_lon_lat_columns(self.table.colnames)

                if "lon_att" not in kwargs:
                    self.lon_att = lon_guess or self.table.colnames[0]

                if "lat_att" not in kwargs:
                    self.lat_att = lat_guess or self.table.colnames[1]

    @validate("coord_type")
    def _check_coord_type(self, proposal):
        if proposal["value"] in ("spherical", "rectangular"):
            return proposal["value"]
        else:
            raise ValueError("coord_type should be spherical or rectangular")

    # Attributes for spherical coordinates

    @validate("lon_unit")
    def _check_lon_unit(self, proposal):
        # Pass the proposal to Unit - this allows us to validate the unit,
        # and allows strings to be passed.
        unit = u.Unit(proposal["value"])
        unit = pick_unit_if_available(unit, VALID_LON_UNITS)
        if unit in VALID_LON_UNITS:
            return unit
        else:
            raise ValueError(
                "lon_unit should be one of {0}".format(
                    "/".join(sorted(str(x) for x in VALID_LON_UNITS))
                )
            )

    @validate("alt_unit")
    def _check_alt_unit(self, proposal):
        # Pass the proposal to Unit - this allows us to validate the unit,
        # and allows strings to be passed.
        with u.imperial.enable():
            unit = u.Unit(proposal["value"])
        unit = pick_unit_if_available(unit, VALID_ALT_UNITS)
        if unit in VALID_ALT_UNITS:
            return unit
        else:
            raise ValueError(
                "alt_unit should be one of {0}".format(
                    "/".join(sorted(str(x) for x in VALID_ALT_UNITS))
                )
            )

    @validate("alt_type")
    def _check_alt_type(self, proposal):
        if proposal["value"] in VALID_ALT_TYPES:
            return proposal["value"]
        else:
            raise ValueError(
                "alt_type should be one of {0}".format(
                    "/".join(str(x) for x in VALID_ALT_TYPES)
                )
            )

    @validate("time_att")
    def _check_time_att(self, proposal):
        # Parse the time_att column and make sure it's in the proper format
        # (string in isot format, astropy Time, or datetime)
        col = self._get_table()[proposal["value"]]

        if all(isinstance(t, datetime) for t in col) or all(
            isinstance(t, Time) for t in col
        ):
            return proposal["value"]

        elif isinstance(col, STR_TYPE) or np.issubdtype(col.dtype, NP_STR_TYPE):

            try:
                Time(col, format="isot")
                return proposal["value"]
            except ValueError:
                raise ValueError(
                    "String times must conform to the ISOT"
                    "standard (YYYY-MM-DD`T`HH:MM:SS:MS)"
                )

        else:
            raise ValueError(
                "A time column must only have string, "
                "datetime.datetime, or astropy Time values"
            )

    @validate("time_decay")
    def _check_decay(self, proposal):
        if proposal["value"].unit.physical_type == "time":
            return proposal["value"]
        else:
            raise ValueError("time_decay should be in units of time")

    @observe("alt_att")
    def _on_alt_att_change(self, *value):
        # Check if we can set the unit of the altitude automatically
        if not self.notify_changes or len(self.alt_att) == 0:
            return
        column = self._get_table()[self.alt_att]
        unit = pick_unit_if_available(column.unit, VALID_ALT_UNITS)
        if unit in VALID_ALT_UNITS:
            self.alt_unit = unit
        elif unit is not None:
            warnings.warn(
                "Column {0} has units of {1} but this is not a valid"
                " unit of altitude - set the unit directly with"
                " alt_unit".format(self.alt_att, unit),
                UserWarning,
            )

    @observe("lon_att")
    def _on_lon_att_change(self, *value):
        # Check if we can set the unit of the longitude automatically
        if not self.notify_changes or len(self.lon_att) == 0:
            return
        column = self._get_table()[self.lon_att]
        unit = pick_unit_if_available(column.unit, VALID_LON_UNITS)
        if unit in VALID_LON_UNITS:
            self.lon_unit = unit
        elif unit is not None:
            warnings.warn(
                "Column {0} has units of {1} but this is not a valid "
                "unit of longitude - set the unit directly with "
                "lon_unit".format(self.lon_att, unit),
                UserWarning,
            )

    # Attributes for cartesian coordinates

    @observe("x_att", "y_att", "z_att")
    def _on_xyz_att_change(self, *value):
        # Check if we can set the unit of the x/y/z positions automatically
        if not self.notify_changes:
            return
        for att in (self.x_att, self.y_att, self.z_att):
            if len(att) == 0:
                continue
            column = self._get_table()[att]
            unit = pick_unit_if_available(column.unit, VALID_ALT_UNITS)
            if unit in VALID_ALT_UNITS:
                self.xyz_unit = unit
                return
            elif unit is not None:
                warnings.warn(
                    "Column {0} has units of {1} but this is not a valid "
                    "unit of distance - set the unit directly with "
                    "xyz_unit".format(self.alt_att, unit),
                    UserWarning,
                )

    @validate("xyz_unit")
    def _check_xyz_unit(self, proposal):
        # Pass the proposal to Unit - this allows us to validate the unit,
        # and allows strings to be passed.
        unit = u.Unit(proposal["value"])
        unit = pick_unit_if_available(unit, VALID_ALT_UNITS)
        if unit in VALID_ALT_UNITS:
            return unit
        else:
            raise ValueError(
                "xyz_unit should be one of {0}".format(
                    "/".join(sorted(str(x) for x in VALID_ALT_UNITS))
                )
            )

    # Visual attributes

    @validate("marker_type")
    def _check_marker_type(self, proposal):
        if proposal["value"] in VALID_MARKER_TYPES:
            return proposal["value"]
        else:
            raise ValueError(
                "marker_type should be one of {0}".format(
                    "/".join(str(x) for x in VALID_MARKER_TYPES)
                )
            )

    @validate("marker_scale")
    def _check_marker_scale(self, proposal):
        if proposal["value"] in VALID_MARKER_SCALES:
            return proposal["value"]
        else:
            raise ValueError(
                "marker_scale should be one of {0}".format(
                    "/".join(str(x) for x in VALID_MARKER_SCALES)
                )
            )

    @validate("cmap")
    def _check_cmap(self, proposal):
        if isinstance(proposal["value"], str):
            return cm.get_cmap(proposal["value"])
        elif not isinstance(proposal["value"], Colormap):
            raise TypeError("cmap should be set to a Matplotlib colormap")
        else:
            return proposal["value"]

    @observe("size_att")
    def _on_size_att_change(self, *value):
        # Set the min/max levels automatically based on the min/max values
        if not self.notify_changes:
            return
        if len(self.size_att) == 0:
            self.parent._send_msg(
                event="table_layer_set", id=self.id, setting="sizeColumn", value=-1
            )
            return

        self.size_vmin = None
        self.size_vmax = None

        column = self._get_table()[self.size_att]

        self.size_vmin = np.nanmin(column)
        self.size_vmax = np.nanmax(column)

    @observe("size_vmin", "size_vmax")
    def _on_size_vmin_vmax_change(self, *value):
        # Update the size column in the table
        if not self.notify_changes:
            return

        if self._uniform_size():
            self.parent._send_msg(
                event="table_layer_set", id=self.id, setting="sizeColumn", value=-1
            )
        else:
            self.parent._send_msg(
                event="table_layer_set", id=self.id, setting="pointScaleType", value=0
            )

            self.parent._send_msg(
                event="table_layer_set",
                id=self.id,
                setting="sizeColumn",
                value=self.size_att,
            )

            self.parent._send_msg(
                event="table_layer_set", id=self.id, setting="normalizeSize", value=True
            )

            self.parent._send_msg(
                event="table_layer_set",
                id=self.id,
                setting="normalizeSizeClip",
                value=True,
            )

            self.parent._send_msg(
                event="table_layer_set",
                id=self.id,
                setting="normalizeSizeMin",
                value=self.size_vmin,
            )

            self.parent._send_msg(
                event="table_layer_set",
                id=self.id,
                setting="normalizeSizeMax",
                value=self.size_vmax,
            )

    @observe("cmap_att")
    def _on_cmap_att_change(self, *value):

        # Set the min/max levels automatically based on the min/max values

        if not self.notify_changes:
            return

        if len(self.cmap_att) == 0:
            self.parent._send_msg(
                event="table_layer_set", id=self.id, setting="colorMapColumn", value=-1
            )

            self.parent._send_msg(
                event="table_layer_set", id=self.id, setting="colorMap", value=0
            )

            return

        self.cmap_vmin = None
        self.cmap_vmax = None

        column = self._get_table()[self.cmap_att]

        self.cmap_vmin = np.nanmin(column)
        self.cmap_vmax = np.nanmax(column)

    @observe("cmap_vmin", "cmap_vmax", "cmap")
    def _on_cmap_vmin_vmax_change(self, *value):

        # Update the cmap column in the table

        if not self.notify_changes:
            return

        if self._uniform_color():

            self.parent._send_msg(
                event="table_layer_set", id=self.id, setting="colorMapColumn", value=-1
            )

            self.parent._send_msg(
                event="table_layer_set", id=self.id, setting="colorMap", value=0
            )

        else:

            self.parent._send_msg(
                event="table_layer_set", id=self.id, setting="colorMap", value=3
            )

            if self.cmap.name.lower() in VALID_COLORMAPS:

                self.parent._send_msg(
                    event="table_layer_set",
                    id=self.id,
                    setting="colorMapColumn",
                    value=self.cmap_att,
                )

                self.parent._send_msg(
                    event="table_layer_set",
                    id=self.id,
                    setting="colorMapperName",
                    value=self.cmap.name,
                )

                self.parent._send_msg(
                    event="table_layer_set",
                    id=self.id,
                    setting="dynamicColor",
                    value=True,
                )

                self.parent._send_msg(
                    event="table_layer_set",
                    id=self.id,
                    setting="normalizeColorMap",
                    value=True,
                )

                self.parent._send_msg(
                    event="table_layer_set",
                    id=self.id,
                    setting="normalizeColorMapMin",
                    value=self.cmap_vmin,
                )

                self.parent._send_msg(
                    event="table_layer_set",
                    id=self.id,
                    setting="normalizeColorMapMax",
                    value=self.cmap_vmax,
                )

            else:
                table = self._get_table()
                column = table[self.cmap_att]

                values = (column - self.cmap_vmin) / (self.cmap_vmax - self.cmap_vmin)

                # PERF: vectorize the calculation of the hex strings
                rgb = self.cmap(values)[:, :-1]
                hex_values = [to_hex(x) for x in rgb]

                table[CMAP_COLUMN_NAME] = hex_values

                self.parent._send_msg(
                    event="table_layer_update", id=self.id, table=self._table_b64
                )

                self.parent._send_msg(
                    event="table_layer_set",
                    id=self.id,
                    setting="dynamicColor",
                    value=False,
                )

                self.parent._send_msg(
                    event="table_layer_set",
                    id=self.id,
                    setting="colorMapColumn",
                    value=CMAP_COLUMN_NAME,
                )

    @observe("time_att")
    def _on_time_att_change(self, *value):

        if (
            not self.notify_changes
            or len(self.time_att) == 0
            or self.time_series is False
        ):
            self.parent._send_msg(
                event="table_layer_set", id=self.id, setting="startDateColumn", value=-1
            )
            return

        wwt_times = Column(self._get_table()[self.time_att].copy()).tolist()
        # must specify Column so we can use tolist() on astropy Time columns

        # Convert time column to UTC so WWT displays points at expected times
        for i, tm in enumerate(wwt_times):
            wwt_times[i] = ensure_utc(tm, str_allowed=True)

        # Update the table passed to WWT with the new, modified time column
        self._get_table()[TIME_COLUMN_NAME] = Column(wwt_times)

        self.parent._send_msg(
            event="table_layer_update", id=self.id, table=self._table_b64
        )

        self.parent._send_msg(
            event="table_layer_set",
            id=self.id,
            setting="startDateColumn",
            value=TIME_COLUMN_NAME,
        )

    def _get_table(self):
        return self.table

    @property
    def _table_b64(self):

        # TODO: We need to make sure that the table has ra/dec columns since
        # WWT absolutely needs that upon creation.

        csv = csv_table_win_newline(self._get_table())

        return b64encode(csv.encode("ascii", errors="replace")).decode("ascii")

    def _uniform_color(self):
        return not self.cmap_att or self.cmap_vmin is None or self.cmap_vmax is None

    def _uniform_size(self):
        return not self.size_att or self.size_vmin is None or self.size_vmax is None

    def _initialize_layer(self):
        self.parent._send_msg(
            event="table_layer_create",
            id=self.id,
            table=self._table_b64,
            frame=self.frame,
        )

    def update_data(self, table=None):
        """
        Update the underlying data.
        """
        self.table = table.copy(copy_data=False)
        self.parent._send_msg(
            event="table_layer_update", id=self.id, table=self._table_b64
        )

        if len(self.alt_att) > 0:
            if self.alt_att in self.table.colnames:
                self._on_alt_att_change()
            else:
                self.alt_att = ""

        lon_guess, lat_guess = guess_lon_lat_columns(self.table.colnames)

        if self.lon_att in self.table.colnames:
            self._on_lon_att_change()
        else:
            self.lon_att = lon_guess or self.table.colnames[0]

        if self.lat_att not in self.table.colnames:
            self.lat_att = lat_guess or self.table.colnames[1]

    def remove(self):
        """
        Remove the layer.
        """
        if self._removed:
            return
        self.parent._send_msg(event="table_layer_remove", id=self.id)
        self._removed = True
        if self._manager is not None:
            self._manager.remove_layer(self)

    def _on_trait_change(self, changed):
        # This method gets called anytime a trait gets changed. Since this
        # class gets inherited by the Jupyter widgets class which adds some
        # traits of its own, we only want to react to changes in traits
        # that have the wwt metadata attribute (which indicates the name of
        # the corresponding WWT setting).
        if not self.notify_changes:
            return

        wwt_name = self.trait_metadata(changed["name"], "wwt")
        if wwt_name is not None:
            value = changed["new"]
            if changed["name"] == "alt_unit":
                value = VALID_ALT_UNITS[self._check_alt_unit({"value": value})]
            elif changed["name"] == "lon_unit":
                value = VALID_LON_UNITS[self._check_lon_unit({"value": value})]
            elif changed["name"] == "xyz_unit":
                value = VALID_ALT_UNITS[self._check_xyz_unit({"value": value})]
            elif changed["name"] == "time_decay":
                value = value.to(u.day).value
            self.parent._send_msg(
                event="table_layer_set", id=self.id, setting=wwt_name, value=value
            )

    def _serialize_state(self):
        state = {
            "id": self.id,
            "layer_type": "table",
            "frame": self.frame,
            "settings": {},
        }

        for trait in self.traits().values():
            wwt_name = trait.metadata.get("wwt")
            if wwt_name:
                value = trait.get(self)
                if wwt_name == "raUnits" and value is not None:
                    value = VALID_LON_UNITS[value]
                elif wwt_name == "altUnit" and value is not None:
                    value = VALID_ALT_UNITS[value]
                state["settings"][wwt_name] = value

        if self._uniform_color():
            state["settings"]["_colorMap"] = 0
            state["settings"]["colorMapColumn"] = -1
        else:
            state["settings"]["_colorMap"] = 3
            state["settings"]["colorMapColumn"] = CMAP_COLUMN_NAME

        if self._uniform_size():
            state["settings"]["sizeColumn"] = -1
        else:
            state["settings"]["sizeColumn"] = SIZE_COLUMN_NAME
            state["settings"]["pointScaleType"] = 0

        return state

    def _save_data_for_serialization(self, dir):
        file_path = path.join(dir, "{0}.csv".format(self.id))
        table_str = csv_table_win_newline(self.table)
        with open(
            file_path, "wb"
        ) as file:  # binary mode to preserve windows line endings
            file.write(table_str.encode("ascii", errors="replace"))

    def __str__(self):
        return "TableLayer with {0} markers".format(len(self.table))

    def __repr__(self):
        return "<{0}>".format(str(self))


class CatalogHipsLayer(TableLayer):
    # Reversing the hack in the baseline TableLayer -- here the sizing is
    # entirely controlled by the engine, so just respect it.

    size_att = Unicode(help="The column to use for the marker size (`str`)").tag(
        wwt="sizeColumn"
    )

    size_vmin = Float(
        None,
        help="The minimum point size. Found automagically once size_att is set (`float`)",
        allow_none=True,
    ).tag(wwt="normalizeSizeMin")

    size_vmax = Float(
        None,
        help="The maximum point size. Found automagically once size_att is set (`float`)",
        allow_none=True,
    ).tag(wwt="normalizeSizeMax")

    _settingsReverseMap = {}
    "Intentional class-wide dict - reverse-maps WWT names to pywwt names"

    def __init__(self, parent, id, app_msg):
        self.parent = parent
        self.id = id

        # Set up the settings reverse map if needed.

        if not self._settingsReverseMap:
            for tname, trait in self.traits().items():
                wwt_name = trait.metadata.get("wwt")

                if wwt_name is not None:
                    self._settingsReverseMap[wwt_name] = tname

        # Synchronize Python state with what the app has told us. One day, pywwt
        # should be able to learn about existing non-HiPS table layers in
        # similar fashion, and do the same thing.

        self.notify_changes = False

        # skeleton table so that we can reverse-map column names:
        self.table = Table(names=app_msg["spreadsheetInfo"]["header"])

        settings = app_msg["spreadsheetInfo"]["settings"]

        def mapCol(x):
            if x == -1:
                return None
            return self.table.colnames[x]

        for wwt_name, val in settings:
            pyname = self._settingsReverseMap.get(wwt_name)
            if pyname is None:
                continue

            if "_att" in pyname:
                val = mapCol(val)
                if val is None:
                    continue
            elif pyname in ("alt_unit", "xyz_unit"):
                for pyunit, wwtunit in VALID_ALT_UNITS.items():
                    if val == wwtunit:
                        val = pyunit
                        break
            elif pyname == "lon_unit":
                for pyunit, wwtunit in VALID_LON_UNITS.items():
                    if val == wwtunit:
                        val = pyunit
                        break
            elif pyname == "time_decay":
                val = val * u.day

            setattr(self, pyname, val)

        # Regular init, except that we don't override the settings.
        super().__init__(
            parent=parent,
            table=self.table,
            frame="Sky",
            id=id,
            table_from_wwt_engine=True,
        )

        # Finally, honor any future settings changes.
        self.notify_changes = True

    def _get_table(self):
        if not len(self.table):
            raise Exception(
                "HiPS catalog table must be refreshed asynchronously: `await table.refresh()`"
            )
        return self.table

    async def refresh(self, timeout=60):
        """
        Update the Python table with the data that are currently visible in the
        WWT viewer.

        Returns
        -------
        table : :class:`~astropy.table.Table`
        """

        fut = self.parent._send_into_future(
            timeout=timeout,
            event="layer_hipscat_datainview",
            tableId=self.id,
            limit=True,
        )

        reply = await fut
        self.table = Table.read(reply["data"], format="ascii.tab")
        return self.table

    def update_data(self, table=None):
        raise Exception(
            "HiPS catalogs data can only be updated by changing the field of view"
        )

    def __str__(self):
        return "Catalog HiPS Layer: {0}".format(self.id)


class ImageLayer(HasTraits):
    """
    An image layer.
    """

    vmin = Float(None, allow_none=True)
    vmax = Float(None, allow_none=True)
    stretch = Unicode("linear")
    opacity = Float(1, help="The opacity of the image").tag(wwt="opacity")
    cmap = Any(
        cm.viridis,
        help="The Matplotlib colormap " "(:class:`matplotlib.colors.ListedColormap`)",
    ).tag(wwt=None)

    def __init__(self, parent=None, image=None, url=None, **kwargs):
        self.parent = parent
        self.id = str(uuid.uuid4())

        # Attribute to keep track of the manager, so that we can notify the
        # manager if a layer is removed.
        self._manager = None
        self._removed = False

        self._stretch_version = 0
        self._cmap_version = 0

        if image is not None:
            self.name = kwargs.pop("name", None)

            if not self.name:
                if isinstance(image, str):
                    file_name = path.basename(image).split(".gz")[0]
                    self.name = file_name[: file_name.rfind(".")]
                else:
                    self.name = self.id

            # "Classic" mode, processing a single FITS-like input. Transform the
            # image so that it is always acceptable to WWT (Equatorial, TAN
            # projection, double values) and write out to a temporary file.
            self._sanitized_image = tempfile.mktemp()
            sanitize_image(image, self._sanitized_image, **kwargs)
            kwargs.pop("hdu_index", None)

            # The first thing we need to do is make sure the image is being served.
            # For now we assume that image is a filename, but we could do more
            # detailed checks and reproject on-the-fly for example.

            self._image_url = self.parent._serve_file(
                self._sanitized_image, extension=".fits"
            )

            data = fits.getdata(self._sanitized_image)
            self.vmin, self.vmax = np.nanpercentile(data, [0.5, 99.5])
            self._data_min = np.nanmin(data)
            self._data_max = np.nanmax(data)

            self.parent._send_msg(
                event="image_layer_create",
                id=self.id,
                url=self._image_url,
                mode="fits",
                name=self.name,
            )
        elif url is not None:
            # Loading image by URL.
            self._sanitized_image = None
            self._image_url = url

            self.name = kwargs.pop("name", None)
            if not self.name:
                self.name = url

            self.parent._send_msg(
                event="image_layer_create",
                id=self.id,
                url=self._image_url,
                mode="preloaded",
                name=self.name,
            )

        # Check that all kwargs are valid -- throws error if not
        validate_traits(self, kwargs)

        super(ImageLayer, self).__init__(**kwargs)

        # Apply settings (that the user may have overridden) and track changes
        # automagically going forward.
        self._on_trait_change({"name": "vmin", "new": self.vmin})
        self._on_trait_change({"name": "opacity", "new": self.opacity})
        self._on_cmap_change()

        self.observe(self._on_trait_change, type="change")

    @validate("stretch")
    def _check_stretch(self, proposal):
        if proposal["value"] in VALID_STRETCHES:
            return proposal["value"]
        else:
            raise ValueError(
                "stretch should be one of {0}".format(
                    "/".join(str(x) for x in VALID_STRETCHES)
                )
            )

    @validate("cmap")
    def _check_cmap(self, proposal):
        if isinstance(proposal["value"], str):
            if proposal["value"] not in VALID_COLORMAPS:
                raise ValueError(
                    "Colormap should be one of "
                    + "/".join(VALID_COLORMAPS)
                    + " (got {0})".format(proposal["value"])
                )
            return cm.get_cmap(proposal["value"])
        elif not isinstance(proposal["value"], Colormap):
            raise TypeError("cmap should be set to a Matplotlib colormap")
        else:
            if proposal["value"].name not in VALID_COLORMAPS:
                raise ValueError(
                    "Colormap should be one of "
                    + ", ".join(VALID_COLORMAPS)
                    + " (got {0})".format(proposal["value"].name)
                )
            return proposal["value"]

    def remove(self):
        """
        Remove the layer.
        """
        if self._removed:
            return
        self.parent._send_msg(event="image_layer_remove", id=self.id)
        self._removed = True
        if self._manager is not None:
            self._manager.remove_layer(self)

    @observe("cmap")
    def _on_cmap_change(self, *value):
        self._cmap_version += 1
        self.parent._send_msg(
            event="image_layer_cmap",
            id=self.id,
            setting="colorMapperName",
            cmap=self.cmap.name,
            version=self._cmap_version,
        )

    def _on_trait_change(self, changed):
        if changed["name"] in ("stretch", "vmin", "vmax"):
            if self.vmin is not None and self.vmax is not None:
                stretch_id = VALID_STRETCHES.index(self.stretch)
                self._stretch_version += 1
                self.parent._send_msg(
                    event="image_layer_stretch",
                    id=self.id,
                    stretch=stretch_id,
                    vmin=self.vmin,
                    vmax=self.vmax,
                    version=self._stretch_version,
                )

        # This method gets called anytime a trait gets changed. Since this
        # class gets inherited by the Jupyter widgets class which adds some
        # traits of its own, we only want to react to changes in traits
        # that have the wwt metadata attribute (which indicates the name of
        # the corresponding WWT setting).
        wwt_name = self.trait_metadata(changed["name"], "wwt")
        if wwt_name is not None:
            self.parent._send_msg(
                event="image_layer_set",
                id=self.id,
                setting=wwt_name,
                value=changed["new"],
            )

    def _serialize_state(self):
        state = {"id": self.id, "layer_type": "image", "settings": {}}

        # A bit overkill for just the opacity, but more future-proof in case
        # we add more wwt traits
        for trait in self.traits().values():
            wwt_name = trait.metadata.get("wwt")
            if wwt_name:
                state["settings"][wwt_name] = trait.get(self)

        if self.vmin is not None and self.vmax is not None:
            state["stretch_info"] = {
                "vmin": self.vmin,
                "vmax": self.vmax,
                "stretch": VALID_STRETCHES.index(self.stretch),
                "cmap": self.cmap.name,
            }

        return state

    def _save_data_for_serialization(self, dir):
        if self._sanitized_image is None:
            raise Exception("cannot serialize non-local imagery")

        file_path = path.join(dir, "{0}.fits".format(self.id))
        shutil.copyfile(self._sanitized_image, file_path)

    def __str__(self):
        return "ImageLayer"

    def __repr__(self):
        return "<{0}>".format(str(self))



# Copyright 2021 the .NET Foundation
# Licensed under the three-clause BSD License

"""
For Qt desktop applications, a global singleton "QApp" instance.
"""

from qtpy import QtWidgets

__all__ = ['get_qapp', 'cleanup_qapp']

app = None


def get_qapp():
    global app
    if app is None:
        app = QtWidgets.QApplication.instance()
        if app is None:
            app = QtWidgets.QApplication([''])
    return app


def cleanup_qapp():
    global app
    if app is not None:
        app.exit()
        app = None



# Copyright 2021 the .NET Foundation
# Licensed under the three-clause BSD License

"""
The toplevel pywwt package contains only a few miscellaneous types.

Almost all usage of pywwt goes through a variable, conventionally named ``wwt``,
that refers to a subclass of the `BaseWWTWidget` class exported by this module.
The way that you obtain this variable depends on the context in which you're
running your Python code:

- In a JupyterLab environment (most recommended), use
  `pywwt.jupyter.connect_to_app`
- In a plain Jupyter notebook, create an instance of
  `pywwt.jupyter.WWTJupyterWidget`
- In a Qt desktop application, create an instance of `pywwt.qt.WWTQtClient`
- To control an instance of the WWT Windows application, use the unmaintained
  `pywwt.windows` module.

"""

# I don't love all of these imports, but they happened historically through
# asterisk imports so we need to keep them. And Sphinx complains if we don't
# put these types in __all__ up here.
from ._version import version_info, __version__  # noqa
from .core import BaseWWTWidget, DataPublishingNotAvailableError, ViewerNotAvailableError  # noqa
from .annotation import Annotation, Circle, FieldOfView, Line, Polygon  # noqa
from .jupyter_server import load_jupyter_server_extension  # noqa


__all__ = [
    "__version__",
    "Annotation",
    "BaseWWTWidget",
    "Circle",
    "DataPublishingNotAvailableError",
    "FieldOfView",
    "Line",
    "Polygon",
    "ViewerNotAvailableError",
    "load_jupyter_server_extension",
    "version_info",
]


def _jupyter_nbextension_paths():
    return [{'section': 'notebook',
             'src': 'nbextension/static',
             'dest': 'pywwt',
             'require': 'pywwt/extension'}]


def _jupyter_server_extension_paths():
    return [{"module": "pywwt"}]



# Copyright 2021 the .NET Foundation
# Licensed under the three-clause BSD License

"""
pywwt client support for the WWT Jupyter Kernel Data Relay.

Say that you're running pywwt inside JupyterLab, and you want to use it to view
a large FITS file. That file needs to be broken into tiles in order for it to be
viewable: it's both unpleasant and impractical to require the your web browser
to download a gigabyte of data in order to render it with WebGL. That's why we
have tools like ``toasty`` to automate our data processing.

But this exposes a data flow problem. The WWT viewer is running in your web
browser. It's talking to a Jupyter HTTP server, possibly across the internet,
which is in turn talking to a Python kernel which in turn may be running on yet
another machine, quite possibly one that's not accessible on the open internet.
In order to view your data, we need to make data files on the pywwt machine
"visible" on the open internet so that WWT, running in your web browser, can
retrieve them.

The only reliable way to do this is through the Jupyter server. Unfortunately we
know of no built-in way for kernels to plug data into the server. Therefore we
have created the ``wwt_kernel_data_relay`` Python package, a Jupyter server
extension that makes it possible for kernels to request to publish data through
the Jupyter server. This publication is done using extensions to the ZMQ-based
messaging framework that Jupyter servers use to control Jupyter kernels.

This module implements the kernel side of the Kernel Data Relay (KDR) protocol
for pywwt. The ultimate interface is that other code can request that individual
files or directory trees be made available for service. This module returns a
partial URL that can then be forwarded to the Jupyter frontend (the JS layer,
not the server) or the user (e.g. by printing it out).
"""

import contextlib
from hashlib import md5
import mimetypes
import os.path
from urllib.parse import quote as urlquote

from ipykernel.kernelbase import Kernel

__all__ = [
    "get_relay_hub",
    "JupyterRelayHub",
]

FIXED_HTTP_RESPONSE_HEADERS = [
    ("Access-Control-Allow-Origin", "*"),
    ("Access-Control-Allow-Methods", "GET,HEAD"),
    (
        "Access-Control-Allow-Headers",
        "Content-Disposition,Content-Encoding,Content-Length,Content-Type",
    ),
]

STATIC_DIR = os.path.join(os.path.dirname(__file__), "web_static")


class HTTPExposedError(Exception):
    """
    Quick exception class for service errors that should be exposed to the user
    as specific HTTP error codes.
    """

    def __init__(self, http_status, message):
        self.http_status = http_status
        self.message = str(message)


def default_path_key(path):
    """
    Generate a default KDR key for filesystem path data.

    This is trickier than you might think. We want keys to be stable: if you
    restart your kernel and rerun your code, the KDR key should be the same, so
    that the WWT frontend can still retrieve data from the same URLs. But keys
    have to be uniquified somehow: if you're running pywwt in two different
    notebooks, with two different kernels, they need to be using different keys
    as much as we can manage.

    Our solution for path-based data is to just hash the path. It will be
    stable, and the only way that we should get two kernels claiming the same
    key is if they're trying to serve the same data.
    """

    return md5(path.encode("utf-8")).hexdigest()


class JupyterRelayHub(object):
    """
    pywwt's WWT Kernel Data Relay client.

    This is a singleton, per-process object, because Jupyter kernel-server
    connections are per-process.
    """

    def __init__(self, kernel=None, key_prefix="pywwt_"):
        if kernel is None:
            kernel = Kernel.instance()

        self._kernel = kernel
        self._key_prefix = key_prefix
        self._key_to_resource = {}
        self._base_url = get_notebook_server_base_url()
        self._static_files_url = None
        self._ident_to_next_seqnum = {}

        kernel.shell_handlers["wwtkdr_resource_request"] = self._handle_resource_request

    def get_static_files_url(self):
        """
        Get a partial URL where static files bundled with pywwt can be obtained.

        Returns
        -------
        url : :class:`str`
            A partial URL.

        Notes
        -----
        The returned URL will not be an absolute URL. It will end with a
        trailing slash. In order to be turned into a retrievable URL, it must be
        combined with the public-facing URL of the running Jupyter server in the
        JavaScript frontend -- the kernel does not have that information.

        The static files are stored in the ``pywwt/web_static`` subdirectory of
        the distributed pywwt package.

        Unlike user data, the static files are served with a predictable KDR key
        based on the version of pywwt, since identical versions should be
        serving identical data.
        """

        if self._static_files_url is None:
            from . import version_info

            vcode = ".".join(str(x) for x in version_info)
            self._static_files_url = self.serve_tree(STATIC_DIR, key="static" + vcode)

        return self._static_files_url

    def _claim_key(self, prefixed_key, resource):
        """
        Claim a "key" used by the Kernel Data Relay to map URLs onto running
        kernels. One kernel (i.e., our process) can claim numerous keys.

        Returns the partial URL associated with the key, ending in a trailing
        slash.
        """
        self._key_to_resource[prefixed_key] = resource

        self._kernel.session.send(
            self._kernel.iopub_socket,
            "wwtkdr_claim_key",
            {"key": prefixed_key},
            parent=self._kernel.get_parent("shell"),
            ident=self._kernel._topic("wwtkdr_claim_key"),
        )

        return "{}wwtkdr/{}/".format(self._base_url, urlquote(prefixed_key, safe=""))

    def serve_tree(self, path, key=None):
        """
        Request that a directory tree of files be published.

        Parameters
        ----------
        path : :class:`str`
            A local filesystem path, pointing at a directory
        key : :class:`str` or ``None`` (the default)
            An optional "key" to uniquify the resulting service URL

        Returns
        -------
        url : :class:`str`
            A partial URL, ending in a trailing slash, below which the file data
            will be made accessible.

        Notes
        -----
        The returned URL will not be an absolute URL. It will end with a
        trailing slash. In order to be turned into a retrievable URL, it must be
        combined with the public-facing URL of the running Jupyter server in the
        JavaScript frontend -- the kernel does not have that information.

        If *key* is unspecified, a stable-but-opaque KDR key is used. Currently
        it's a cryptographic hash of the input path.
        """

        if key is None:
            key = default_path_key(path)

        prefixed_key = self._key_prefix + key
        return self._claim_key(prefixed_key, FileTreeResource(path, public=True))

    def serve_file(self, path, extension="", key=None):
        """
        Request that a single specific file be published.

        Parameters
        ----------
        path : :class:`str`
            A local filesystem path, pointing at a non-directory
        extension : :class:`str`, defaults to ``""``
            A file extension to append to the exposed URL
        key : :class:`str` or ``None`` (the default)
            An optional "key" to uniquify the resulting service URL

        Returns
        -------
        url : :class:`str`
            A partial URL where the file will be published
        """
        """
        Notes
        -----
        The returned URL will not be an absolute URL. It will have a structure
        resembling: ``{stuff...}/{key}/{basename(path)}{extension}``. In order
        to be turned into a retrievable URL, it must be combined with the
        public-facing URL of the running Jupyter server in the JavaScript
        frontend -- the kernel does not have that information.

        The *extension* argument is used when the actual filesystem path of the
        data has a filename extension that is missing or misleading for the
        actual data.

        If *key* is unspecified, a stable-but-opaque KDR key is used. Currently
        it's a cryptographic hash of the input path.
        """

        if key is None:
            key = default_path_key(path)

        basename = os.path.basename(path) + extension
        sfr = SingleFileResource(path, basename, public=True)

        prefixed_key = self._key_prefix + key
        url = self._claim_key(prefixed_key, sfr)
        return url + urlquote(basename)

    def _handle_resource_request(self, stream, ident, message):
        """
        This callback is invoked when a ``wwtkdr_resource_request`` message
        arrives on the kernel's "shell" ZMQ stream. It wraps the real
        implementation in some high-level exception handling.
        """

        try:
            self._handle_resource_request_inner(stream, ident, message)
        except HTTPExposedError as e:
            # Some browsers can't "see" cross-origin 404 errors unless the CORS
            # headers are present, so add them here. Note that this can
            # therefore leak some info about the kernel.
            content = {
                "status": "ok",  # this is the correct semantics
                "more": False,
                "http_status": e.http_status,
                "http_headers": [
                    ("Content-Type", "text/plain; charset=utf-8"),
                ]
                + FIXED_HTTP_RESPONSE_HEADERS,
            }

            buffers = [
                memoryview(e.message.encode("utf-8")),
            ]

            self._send_resource_reply(stream, content, message, ident, buffers)
        except Exception as e:
            content = {
                "status": "error",
                "evalue": str(e),
            }

            self._send_resource_reply(stream, content, message, ident, None)
        finally:
            try:
                key = b"".join(ident)
                del self._ident_to_next_seqnum[key]
            except KeyError:
                pass

    def _handle_resource_request_inner(self, stream, ident, message):
        """
        Handling of KDR resource-request messages.

        This is basically equivalent to an HTTP GET request that we have to
        dispatch.
        """

        content = message["content"]

        method = content.get("method", "unspecified")
        if method != "GET":
            raise HTTPExposedError(400, f"unexpected request method {method}")

        key = content.get("key")
        if key is None:
            raise HTTPExposedError(500, "no key in request")

        resource = self._key_to_resource.get(str(key))
        if resource is None:
            raise HTTPExposedError(404, "no such resource registered with kernel")

        # Note that on the server side, Tornado does some normalizations on the
        # input URL before it reaches the KDR handler; for instance,
        # `foo/../bar` is turned into `bar`. But other potentially-sketchy
        # constructs aren't normalized, and the KDR itself applies no
        # validation, so we should still treat `entry` cautiously.
        entry = content.get("entry")
        if entry is None:
            raise HTTPExposedError(500, "no entry in request")

        # The URL experiences the same normalizations that apply to `entry`.
        url = content.get("url")
        if url is None:
            raise HTTPExposedError(500, "no url in request")

        authenticated = content.get("authenticated", False)

        # OK, we've got everything we need from the request. Delegate to the
        # resource.
        handle, headers = resource.entry_to_handle(authenticated, str(entry), str(url))

        with contextlib.closing(handle) as handle:
            first = True
            keep_going = True

            while keep_going:
                # There is some kind of message limit (max 2000 at the same time) on the receiving end.
                # So we want so send as few messages as possible to never reach that limit.
                # Even if the messages are sent correctly, only the first 2000 are received by KDR.
                # We cannot just make sure that every single request consists of less than 2000 messages,
                # since other requests seem to be fighting for the same 2000 spots.
                # So I felt like it is safest to up the chunk size to something very, very high.
                # At least until the root issue is found and resolved.
                chunk = handle.read(8388608)  # 8MB

                if len(chunk):
                    buffers = [chunk]
                else:
                    keep_going = False
                    buffers = []

                content = {
                    "status": "ok",
                    "more": keep_going,
                }

                if first:
                    content["http_status"] = 200
                    content["http_headers"] = headers
                    first = False

                self._send_resource_reply(stream, content, message, ident, buffers)

    def _send_resource_reply(self, stream, content, parent, ident, buffers):
        """
        Send a resource reply message. We centralize this in order to make sure
        that we always send an appropriate sequence number for each message, in
        the face of exceptions that can crop up at surprising times.
        """
        key = b"".join(ident)
        seqnum = self._ident_to_next_seqnum.get(key, 0)
        content["seq"] = seqnum
        self._ident_to_next_seqnum[key] = seqnum + 1

        self._kernel.session.send(
            stream,
            "wwtkdr_resource_reply",
            content=content,
            parent=parent,
            ident=ident,
            buffers=buffers,
        )


def _open_file_resource(path, path_for_mime=None):
    """
    Core implementation of serving up filesystem files.

    The main wrinkles here are mapping ENOENT into 404 errors, and
    determining a Content-Type header.
    """

    # TODO: don't give full CORS headers if resource is not public

    try:
        handle = open(path, "rb")
    except FileNotFoundError:
        raise HTTPExposedError(404, "file not found")

    # Note that the "encoding" here is something like "gzip", not "UTF-8" --
    # it's about the file stream, not text.
    if path_for_mime is None:
        path_for_mime = path

    content_type, _content_encoding = mimetypes.guess_type(path_for_mime)
    if content_type is None:
        content_type = "application/octet-stream"

    headers = FIXED_HTTP_RESPONSE_HEADERS + [
        ("Content-Type", content_type),
    ]
    return handle, headers


def _generate_absolute_wtml(rel_wtml_path, url):
    """
    Generate WTML with absolute URLs on the fly.

    This is about the only piece of this module that's WWT-specific. The same as
    ``wwtdatatool serve``, when making tiled data available through a WTML file,
    we need to fill in the WTML with absolute URLs on the fly, since the tiling
    code can't (reasonably) know what final URL the data will be served at.
    """

    from io import BytesIO
    from wwt_data_formats.folder import Folder, make_absolutizing_url_mutator

    f = Folder.from_file(rel_wtml_path)
    f.mutate_urls(make_absolutizing_url_mutator(url))
    resp = f.to_xml_string().encode("utf-8")
    handle = BytesIO(resp)

    headers = FIXED_HTTP_RESPONSE_HEADERS + [
        ("Content-Type", "application/x-wtml"),
    ]
    return handle, headers


class FileTreeResource(object):
    """
    Internal helper class for serving up a directory tree of files, such as you
    might get after tiling a large image.
    """

    def __init__(self, root, public=False):
        self._root = root
        self._public = public
        assert public, "non-public resources not implemented"

    def entry_to_handle(self, authenticated, entry, url):
        # Lame-brained indexing support:
        if entry.endswith("/"):
            entry += "index.html"

        items = entry.split("/")

        if any(e == ".." or e.startswith("/") or not e for e in items):
            raise HTTPExposedError(400, "illegal kernel data tree path component")

        # Special handling for `index_rel.wtml` files that make tiled data
        # available to the WWT frontend. We need to source `index.wtml` on the
        # fly in order to fill the file with absolute URL data, which the WWT
        # engine requires.
        if entry.endswith(".wtml"):
            rel_path = os.path.join(self._root, *items)[:-5] + "_rel.wtml"
            if os.path.exists(rel_path):
                return _generate_absolute_wtml(rel_path, url)

        return _open_file_resource(os.path.join(self._root, *items))


class SingleFileResource(object):
    """
    Internal helper class for serving up a single specific file.
    """

    def __init__(self, path, url_basename, public=False):
        self._path = path
        self._url_basename = url_basename
        self._public = public
        assert public, "non-public resources not implemented"

    def entry_to_handle(self, authenticated, entry, _url):
        # TODO: handle non-public resources!!!

        if entry != self._url_basename:
            raise HTTPExposedError(404, "file not found")

        # Use the URL basename here in case the `extension` argument was used
        # (e.g. the extension of the actual filesystem path is missing or
        # misleading).
        return _open_file_resource(self._path, path_for_mime=self._url_basename)


_global_relay_hub = None


def get_relay_hub(kernel=None):
    """
    Get the global singleton Kernel Data Relay (KDR) client.

    Returns
    -------
    hub : :class:`JupyterRelayHub`
    """

    global _global_relay_hub

    if _global_relay_hub is None:
        # This is a bit out-of-place, but whatever
        mimetypes.add_type("image/fits", ".fits")
        mimetypes.add_type("image/fits", ".fts")
        mimetypes.add_type("image/fits", ".fit")
        mimetypes.add_type("application/x-wtml", ".wtml")

        _global_relay_hub = JupyterRelayHub(kernel=kernel)

    return _global_relay_hub


# January 2021: Derived from notebook.notebookapp.list_running_servers, with a
# fix for JupyterLab 3.x (or something recent??), where the JSON files start
# with `jpserver` not `nbserver`. This function is part of the infrastructure
# needed to get the notebook server base url.
def _list_running_servers_jl3():
    import io
    import json
    from notebook.utils import check_pid
    from jupyter_core.paths import jupyter_runtime_dir
    import os.path
    import re

    runtime_dir = jupyter_runtime_dir()

    if not os.path.isdir(runtime_dir):
        return

    for file_name in os.listdir(runtime_dir):
        # here is the fix:
        if re.match("nbserver-(.+).json", file_name) or re.match(
            "jpserver-(.+).json", file_name
        ):
            with io.open(os.path.join(runtime_dir, file_name), encoding="utf-8") as f:
                info = json.load(f)

            if ("pid" in info) and check_pid(info["pid"]):
                yield info
            else:
                try:
                    os.unlink(os.path.join(runtime_dir, file_name))
                except OSError:
                    pass


def _compute_notebook_server_base_url():
    """
    Figure out the base_url of the current Jupyter notebook server.

    Copied from
    https://github.com/jupyter/notebook/issues/3156#issuecomment-401119433
    with miniscule changes. This is gross, but appears to be the best
    available option right now.

    """
    import ipykernel
    import json
    import re
    import requests

    # First, find our ID.
    kernel_id = re.search(
        "kernel-(.*).json", ipykernel.connect.get_connection_file()
    ).group(1)

    # Now, check all of the running servers known on this machine. We have to
    # talk to each server to figure out if it's ours or somebody else's.
    running_server_info = list(_list_running_servers_jl3())

    for s in running_server_info:
        # We need an API token that in most cases is provided in the runtime
        # JSON files. In (recent versions of?) the JupyterHub single-user
        # server, it seems that the token is instead obtained from an
        # environment variable. Cf.
        # https://github.com/jupyterhub/jupyterhub/blob/master/jupyterhub/singleuser/mixins.py
        token = s.get("token", "")
        if not token:
            token = os.environ.get("JUPYTERHUB_API_TOKEN", "")
        if not token:
            token = os.environ.get("JPY_API_TOKEN", "")  # deprecated as of 0.7.2

        # Request/response paranoia due to "fun" figuring out how to fix the
        # JupyterHub single-user problem - the API call would fail due to auth
        # issues and break pywwt, even though there was only one running server
        # so we actually didn't even need the API call. In case something breaks
        # in the future, add a fallback mode.
        try:
            response = requests.get(
                requests.compat.urljoin(s["url"], "api/sessions"),
                params={"token": token},
            )

            for n in json.loads(response.text):
                if n["kernel"]["id"] == kernel_id:
                    return s["base_url"]  # Found it!
        except Exception:
            pass

    # If we got here, we might have auth issues with the api/sessions request.
    # If there's only one server, just give it a try.
    if len(running_server_info) == 1:
        return running_server_info[0]["base_url"]

    raise Exception(
        "cannot locate our notebook server; is this code running in a Jupyter kernel?"
    )


_server_base_url = None


def get_notebook_server_base_url():
    """
    Get the "base_url" of the current Jupyter notebook server.

    """
    global _server_base_url

    if _server_base_url is None:
        _server_base_url = _compute_notebook_server_base_url()
        if not _server_base_url.endswith("/"):
            _server_base_url = _server_base_url + "/"

    return _server_base_url



# Copyright 2018-2021 the .NET Foundation
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This package contains the pywwt tests. This particular module contains some
infrastructure to help with the image-output tests, which can be pretty finicky.
"""

__all__ = ['assert_widget_image', 'wait_for_test']

from glob import glob
import os.path
import sys
import tempfile

from matplotlib.testing.compare import compare_images  # noqa
import pytest

DATA = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data'))
IMAGE_OUTPUT_DIR = os.environ.get('PYWWT_TEST_IMAGE_DIR')
IMAGE_COMPARISON_TOLERANCE = 1.6


def assert_widget_image(tmpdir, widget, filename, fail_now=True):
    """
    Render an image from the given WWT widget and assert that it matches an
    expected version. The expected version might vary depending on the platform
    and/or OpenGL renderer, so we allow for multiple references and just check
    if *any* of them matches well enough.

    We used to have specific test images identified for each platform, but the
    bookkeeping for that gets very finicky.
    """

    # If requested, save the "actual" images in another directory that will be
    # preserved beyond the test run.

    if IMAGE_OUTPUT_DIR:
        actual = os.path.join(IMAGE_OUTPUT_DIR, filename)
    else:
        actual = tmpdir.join(filename).strpath

    widget.render(actual)

    # Compare to the references

    refdir = os.path.join(DATA, 'refimg_' + os.path.splitext(filename)[0])
    results = []

    for refbase in sorted(os.listdir(refdir)):
        refname = os.path.splitext(refbase)[0]
        expected = os.path.join(refdir, refbase)
        rv = compare_images(
            expected,
            actual,
            tol=IMAGE_COMPARISON_TOLERANCE,
            in_decorator=True
        )

        if rv is None:
            # Success! Clean up any fail images (mostly for the IMAGE_OUTPUT_DIR mode)
            for p in glob(actual.replace('.png', '_vs_*.png')):
                os.unlink(p)
            return None

        failpath = actual.replace('.png', '-failed-diff.png')
        newfailpath = actual.replace('.png', '_vs_%s.png' % refname)
        os.rename(failpath, newfailpath)
        results.append((refname, rv['rms']))

    # Nothing was good enough :-(
    #
    # We used to have machinery here to emit a "reproduction script" that
    # printed out Python code to recreate the image files using big
    # BASE64-encoded strings, but now we can just use Azure Pipelines artifacts.
    # Consult the Git history if the reproduction script stuff is needed again.

    msg = (
        'observed image %s did not match any references to required RMS tolerance of '
        '%.2f; results were: %s'
    ) % (actual, IMAGE_COMPARISON_TOLERANCE, ', '.join('%s=%.2f' % t for t in results))

    if fail_now:
        pytest.fail(msg, pytrace=False)

    return '{}: {}'.format(filename, msg)


def wait_for_test(wwt, timeout, for_render=False):
    """
    On macOS with software OpenGL, every so often a single call to
    `app.processEvents()` will take 15-30 seconds to execute. In order for the
    tests like `test_full` to properly update the WWT rendering, it seems that
    *several* such "long" iterations have to occur -- something happens during
    those iterations that doesn't happen during the short ones. So, this
    function hacks MacOS to ensure that we iterate for a *long* time and that we
    execute a nontrivial number of event loop iterations. Note that `wwt.wait()`
    will not necessarily run many mainloop iterations, if one of them takes a
    long time.

    It would be nice to know why this is happening and fix the root cause!
    Preliminary investigation does not yield any clues.

    """
    from time import time
    from ..app import get_qapp
    MIN_ITERS = 128
    ALWAYS_EXTRA_LONG = sys.platform.startswith('darwin')

    if for_render and ALWAYS_EXTRA_LONG:
        timeout = 90

    app = get_qapp()
    t0 = time()
    iters = 0

    # Iterate for *at least* MIN_ITERS and *at least* `timeout` seconds.

    for _ in range(MIN_ITERS):
        iters += 1
        app.processEvents()

    while time() - t0 < timeout:
        iters += 1
        app.processEvents()

    dt = time() - t0

    if for_render:
        print(f'wait_for_test: iters={iters} dt={dt} timeout={timeout} always={ALWAYS_EXTRA_LONG}')


def analyze_test_image_corpus(args):
    """
    This function analyzes the corpus of test images to ultimately report
    whether references images need to be added or removed.

    The arguments to the function should be sets of imagery that should pass the
    test suite. Each argument should be a path to either (1) a Zip file
    downloaded from the Azure Pipelines test suite artifacts or (2) a directory
    containing images output from the test suite (saved by setting the
    $PYWWT_TEST_IMAGE_DIR). The latter mode is useful because local developers'
    machines don't always generate imagery that fully matches with what's
    created in the CI system, and we want the test suite to pass for them, so we
    shouldn't restrict ourselves to *only* outputs created by the CI.

    Update the toplevel README.md appropriately if the behavior here is changed.
    """

    # First: index the existing corpus

    name_to_refs = {}

    for refdir in glob(os.path.join(DATA, 'refimg_*')):
        name = os.path.basename(refdir).replace('refimg_', '')
        refs = {}

        for p in os.listdir(refdir):
            if not p.endswith('.png'):
                continue

            refs[p.replace('.png', '')] = set()

        name_to_refs[name] = refs

    # Now trawl the samples and see which refimgs we hit

    def directory_to_paths(dir_path):
        for filename in os.listdir(dir_path):
            name = filename.replace('.png', '')
            if name in name_to_refs:
                yield name, os.path.join(dir_path, filename)

    def zip_to_paths(zip_path):
        from zipfile import ZipFile

        with tempfile.TemporaryDirectory() as tmpdir, ZipFile(zip_path) as zip:
            for zipname in zip.namelist():
                zip.extract(zipname, tmpdir)
                name = os.path.basename(zipname).replace('.png', '')
                if name in name_to_refs:
                    yield name, os.path.join(tmpdir, zipname)

    for path in args:
        if os.path.isdir(path):
            paths = directory_to_paths(path)
            sampname = os.path.basename(path)
        elif path.endswith('.zip'):
            paths = zip_to_paths(path)
            sampname = os.path.basename(path).replace('.zip', '')
        else:
            raise Exception(f'don\'t know how to handle input path `{path}`')

        for name, imgpath in paths:
            refs = name_to_refs[name]
            refnames = sorted(refs.keys())
            found_it = False
            results = []

            for refname in refnames:
                refpath = os.path.join(DATA, 'refimg_' + name, refname + '.png')
                rv = compare_images(
                    refpath,
                    imgpath,
                    tol=IMAGE_COMPARISON_TOLERANCE,
                    in_decorator=True
                )

                if rv is None:
                    refs[refname].add(sampname)
                    found_it = True
                    break

                failpath = imgpath.replace('.png', '-failed-diff.png')
                os.unlink(failpath)
                results.append((refname, rv['rms']))

            if not found_it:
                print(f'no refimg found for {sampname}::{name}:', ', '.join('%s=%.2f' % t for t in results))

    # Now report

    print()
    print('Report:')
    any_reports = False

    for name in sorted(name_to_refs.keys()):
        refs = name_to_refs[name]

        for refname in sorted(refs.keys()):
            refdata = refs[refname]

            if not refdata:
                print(f'- no hits to {name}::{refname}')
                any_reports = True

    if not any_reports:
        print('- no suggested modifications')



# Copyright 2018-2021 the .NET Foundation
# Licensed under the three-clause BSD license

"""
The core WWT widget class.

The most important definition provided in this module is the `BaseWWTWidget`
type, which defines the generic interface for controlling WWT viewers across the
different backends supported by pywwt.
"""

import asyncio
import json
import os
import shutil
import tempfile

from astropy import units as u
from astropy.time import Time
from astropy.coordinates import SkyCoord
import numpy as np
from traitlets import HasTraits, observe, validate, TraitError
import nest_asyncio

from .annotation import Circle, Polygon, Line, FieldOfView, CircleCollection
from .imagery import get_imagery_layers, ImageryLayers
from .instruments import Instruments
from .layers import LayerManager
from .logger import logger
from .solar_system import SolarSystem
from .traits import Color, Bool, Float, Unicode, AstropyQuantity
from .utils import ensure_utc

__all__ = [
    "BaseWWTWidget",
    "DataPublishingNotAvailableError",
    "ViewerNotAvailableError",
]

DEFAULT_SURVEYS_URL = "https://worldwidetelescope.github.io/pywwt/surveys.xml"

VIEW_MODES_2D = [
    "sky",
    "sun",
    "mercury",
    "venus",
    "earth",
    "moon",
    "mars",
    "jupiter",
    "callisto",
    "europa",
    "ganymede",
    "io",
    "saturn",
    "uranus",
    "neptune",
    "pluto",
    "panorama",
]

VIEW_MODES_3D = [
    "solar system",
    "milky way",
    "universe",
]

R2D = 180 / np.pi
R2H = 12 / np.pi


class DataPublishingNotAvailableError(Exception):
    """
    Raised if data need to be published, but publishing service isn't available.
    """

    def __init__(self, msg=None):
        if msg is None:
            msg = "there is no mechanism available for pywwt to publish data to the WWT frontend"
        super(DataPublishingNotAvailableError, self).__init__(msg)


class ViewerNotAvailableError(Exception):
    """
    Raised if an operation needs to communicate with the WWT viewer (app), but
    the app doesn't seem to be responding to messages.
    """

    def __init__(self, msg=None):
        if msg is None:
            msg = (
                "cannot complete the operation because the WWT viewer isn't responding"
            )
        super(ViewerNotAvailableError, self).__init__(msg)


class BaseWWTWidget(HasTraits):
    """
    A class that can control a WWT viewer from Python.

    You should not create instances of this class directly. Instead, you should
    obtain obtain an instance through one of the following mechanisms, depending
    on the environment in which you're running Python and WWT:

    - In JupyterLab (the recommended approach), use
      :func:`pywwt.jupyter.connect_to_app`. This requires that you have
      :ref:`set up the WWT JupyterLab extension <setup-jupyterlab>`.
    - In plain Jupyter, or if you want to use WWT in its widget mode, create an
      instance of :class:`pywwt.jupyter.WWTJupyterWidget`. This requires that
      you have :ref:`set up the pywwt Jupyter widget <setup-jupyter-widget>`.
    - If using Qt, create an instance of :class:`pywwt.qt.WWTQtClient`.
    - If you want to control the WWT Windows application, see the
      :mod:`pywwt.windows` module.

    **Attributes:**

    If you modify most of the attributes listed below, your changes will
    automatically be reflected in the WWT view. However, there are a few special
    attributes that collect important pieces of WWT functionality:

    - :attr:`layers` (:class:`pywwt.layers.LayerManager`) allows you to add new
      data layers to the view.
    - :attr:`solar_system` (:class:`pywwt.solar_system.SolarSystem`) provides
      controls related to WWT's 3D "solar system" mode (which can display much
      more than just the solar system!)

    The following attributes collect information about data and modes that are
    available in the connected WWT viewer:

    - :attr:`available_layers` (list of ``str``) lists available image sets in a
      flat array
    - :attr:`available_views` (list of ``str``) lists available view modes
      (``"solar system"``, ``"jupiter"``, etc.) in a flat array
    - :attr:`imagery` (:class:`pywwt.imagery.ImageryLayers`) lists available
      imagery
    - :attr:`instruments` (:class:`pywwt.instruments.Instruments`) lists available
      instrument FOVs
    """

    _startupMessageQueue = None
    _appAlive = False
    _readyFuture = None
    _readinessAchieved = False

    # View state that the frontend sends to us:
    _raRad = 0.0
    _decRad = 0.0
    _fovDeg = 60.0
    _engineTime = Time("2017-03-09T12:30:00", format="isot")
    _systemTime = Time("2017-03-09T12:30:00", format="isot")
    _timeRate = 1.0

    def __init__(self, hide_all_chrome=False):
        """
        (Note that this docstring is not exposed in the API docs. It is aimed at
        developers.)

        Parameters
        ----------
        hide_all_chrome : optional `bool`
            Configures the WWT frontend to hide all user-interface "chrome".
            Defaults to true to maintain compatibility with the historical
            pywwt user experience.

        Notes
        -----

        This class communicates with the WWT "research application"
        (@wwtelescope/research-app), which has a standardized control interface
        (@wwtelescope/research-app-messages). This client is fundamentally
        concerned with communicating using that messaging interface.

        Subclasses need to set up some kind of mechanism that will call
        ``_on_app_status_change`` when the app is ready to receive messages (at
        a minimum), and ``_on_app_message_received`` when a message from the
        frontend app is received. They need to implement ``_actually_send_msg``
        which will deliver a message to the app.
        """

        super(BaseWWTWidget, self).__init__()

        self.observe(self._on_trait_change, type="change")

        self._startupMessageQueue = []
        self._seqNum = 0
        self._futures = {}

        self._available_layers = get_imagery_layers(DEFAULT_SURVEYS_URL)
        self._available_hips_catalog_names = []
        self.imagery = ImageryLayers(self._available_layers)
        self.solar_system = SolarSystem(self)
        self._instruments = Instruments()
        self.current_mode = "sky"
        self._paused = False
        self._last_sent_view_mode = "sky"
        self.layers = LayerManager(parent=self)
        self._annotation_set = set()
        self._callbacks = {}

        if hide_all_chrome:
            self._send_msg(
                event="modify_settings",
                target="app",
                settings=[
                    ("hideAllChrome", True),
                ],
            )

        # pywwt's surveys.xml has slightly different contents than
        # builtin-image-sets.wtml (for the time being), so we want to make sure
        # that the frontend agrees with us about which named imagesets are
        # available.
        self._send_msg(
            event="load_image_collection",
            url=DEFAULT_SURVEYS_URL,
            threadId=self._next_seq(),
        )

        # Forcibly set up the app's configuration to match the pywwt default.
        # Most things align, but not everything. Note that fg/bg imageset names
        # can't rely on items that are unique to DEFAULT_SURVEYS_URL since it
        # won't necessarily have loaded by the time these messages reach the
        # app.
        #
        # Note that this implementation is bad because every time you create
        # a new client, you'll forcible modify the app's state. We need to
        # move to a system where this initializer can reliably sync its state
        # to the app's.
        self._send_msg(
            event="set_background_by_name",
            name=self.background,
        )

        self._send_msg(
            event="set_foreground_by_name",
            name=self.foreground,
        )

        self._send_msg(
            event="set_foreground_opacity",
            value=self.foreground_opacity * 100,
        )

        SETTINGS = [
            "actual_planet_scale",
            "constellation_boundary_color",
            "constellation_figure_color",
            "constellation_selection_color",
        ]

        for s in SETTINGS:
            wwt_name = self.trait_metadata(s, "wwt")
            self._send_msg(
                event="setting_set",
                setting=wwt_name,
                value=getattr(self, s),
            )

        # NOTE: we deliberately don't force _on_trait_change to be called here
        # for the WWT settings, as the default values are hard-coded in the
        # widget HTML. This is done because there is otherwise no reliable way
        # of making sure that we would call _on_trait_change once WWT is ready
        # to receive commands. There is a test in test_core.py that ensures that
        # the defaults here are in sync with the defaults in the widget HTML.

    def _on_trait_change(self, changed):
        # This method gets called anytime a trait gets changed. Since this class
        # gets inherited by the Jupyter widgets class which adds some traits of
        # its own, we only want to react to changes in traits that have the wwt
        # metadata attribute (which indicates the name of the corresponding WWT
        # setting).
        wwt_name = self.trait_metadata(changed["name"], "wwt")
        new_value = changed["new"]

        if wwt_name is not None:
            if isinstance(new_value, u.Quantity):
                new_value = new_value.value

            self._send_msg(event="setting_set", setting=wwt_name, value=new_value)

    def _next_seq(self):
        """
        When we send requests to the app, we have to be mindful that replies
        will arrive asynchronously. This means that we need to have some kind of
        unique identifier for each "conversation" helping us understand their
        individual progression. This field is called the "threadId" in the
        research app messaging API.

        To generate these unique IDs, we use a simple sequence number as needed.
        The underlying message transport implementation should further uniquify
        these IDs if needed -- this will depend on the message transport
        mechanism.
        """
        self._seqNum += 1
        return str(self._seqNum)

    def _send_msg(self, **kwargs):
        if self._startupMessageQueue is not None:
            self._startupMessageQueue.append(kwargs)
        elif self._appAlive:
            self._actually_send_msg(kwargs)
        else:
            raise ViewerNotAvailableError()

    def _send_into_future(self, timeout=30, **kwargs):
        """
        Send a message and return an asyncio Future that will resolve when the
        message receives a reply from the app. The value of the future will be
        the full JSON message received.

        This interface leverages the WWT messaging convention that messages
        which receive replies will have a field named `threadId` that will be
        reproduced in the reply. The message handler looks for this threadId and
        resolves the future if/when the reply is seen.

        By default, the future will timeout eventually. The timeout parameter is
        measured in seconds. If it's ``None``, no timeout will be applied.
        """

        seq = self._next_seq()
        loop = asyncio.get_running_loop()
        fut = loop.create_future()

        self._futures[seq] = fut
        self._send_msg(threadId=seq, **kwargs)

        if timeout is not None:

            def maybe_time_it_out():
                if not fut.done():
                    fut.set_exception(asyncio.TimeoutError())
                    del self._futures[seq]

            loop.call_later(timeout, maybe_time_it_out)

        return fut

    def _on_app_status_change(self, alive=None):
        """
        Extensibility API: if a keyword is None, that means no change in status.

        This function should be prepared to handle redundant "status change"
        updates because that's the just kind of thing that's going to happen in
        this asynchronous messaging environment.
        """

        if alive is not None:
            self._appAlive = alive

            if alive and self._readyFuture and not self._readinessAchieved:
                # See becomes_ready()
                self._readyFuture.set_result(self)
                self._readinessAchieved = True

            if alive and self._startupMessageQueue:
                queue = self._startupMessageQueue
                self._startupMessageQueue = None

                for msg in queue:
                    self._actually_send_msg(msg)

    def _on_app_message_received(self, payload):
        """
        Call this function when a message is received from the research app.
        This will generally happen in some kind of asynchronous event handler,
        so there is no guarantee that exceptions raised here will be exposed to
        the user.
        """

        ptype = payload.get("type")
        # some events don't have type but do have: pevent = payload.get('event')

        updated_fields = []

        if ptype == "wwt_view_state":
            try:
                self._raRad = float(payload["raRad"])
                self._decRad = float(payload["decRad"])
                self._fovDeg = float(payload["fovDeg"])
                self._engineTime = Time(payload["engineClockISOT"], format="isot")
                self._systemTime = Time(payload["systemClockISOT"], format="isot")
                self._timeRate = float(payload["engineClockRateFactor"])
            except ValueError:
                pass  # report a warning somehow?
        elif ptype == "wwt_application_state":
            hipscat = payload.get("hipsCatalogNames")

            if hipscat is not None:
                self._available_hips_catalog_names = hipscat

        elif ptype == "wwt_selection_state":
            most_recent = payload.get("mostRecentSource")
            sources = payload.get("selectedSources")

            if most_recent is not None:
                self._most_recent_source = most_recent
                updated_fields.append("most_recent_source")

            if sources is not None:
                self._selected_sources = sources
                updated_fields.append("selected_sources")

        # Any relevant async future to resolve?

        tid = payload.get("threadId")

        if tid is not None:
            try:
                fut = self._futures.pop(tid)
            except KeyError:
                pass
            else:
                fut.set_result(payload)

        # Any client-side callbacks to execute?

        callback = self._callbacks.get(ptype)
        if callback:
            try:
                callback(self, updated_fields)
            except:  # noqa: E722
                logger.exception("unhandled Python exception during a callback")

    def _set_message_type_callback(self, ptype, callback):
        """
        Set a callback function that will be executed when the widget receives a message with the given type.

        Parameters
        ----------
        ptype: str
            The string that identifies the message type
        callback: `BaseWWTWidget`
            A callable object which takes two arguments: the WWT widget instance, and a list of updated properties.
        """
        self._callbacks[ptype] = callback

    def set_selection_change_callback(self, callback):
        """
        Set a callback function that will be executed when the widget receives a selection change message.

        Parameters
        ----------
        callback:
            A callable object which takes two arguments: the WWT widget instance, and a list of updated properties.
        """
        self._set_message_type_callback("wwt_selection_state", callback)

    def _get_view_data(self, field):
        if not self._appAlive:
            raise ViewerNotAvailableError()

        if field == "ra":
            return self._raRad * R2H
        elif field == "dec":
            return self._decRad * R2D
        elif field == "fov":
            return self._fovDeg
        elif field == "datetime":
            engine_delta = self._timeRate * (Time.now() - self._systemTime)
            return self._engineTime + engine_delta
        else:
            raise ValueError('internal problem: unexpected "field" value')

    # Support methods that can/should be overridden by subclasses:

    def _actually_send_msg(self, payload):
        """
        Note that the API here is different than ``_send_msg``: we take a dict,
        not ``**kwargs``.
        """
        raise NotImplementedError()

    def _serve_file(self, filename, extension=""):
        """
        Publish a single file in a web server, for use by the WWT frontend.

        Parameters
        ----------
        filename : :class:`str`
            The filesystem path of the data file
        extension : optional :class:`str`, default ``""``
            A custom filename extension to be included in the published URL.

        Returns
        -------
        url : :class:`str`
            A URL, possibly incomplete, at which the file is available.

        Raises
        ------
        :exc:`DataPublishingNotAvailableError`
            Raised if no data publishing service is available.

        Notes
        -----
        Because the WWT engine is a web application, the only way to get data
        into it is to make those data available over HTTP. This method provides
        user Python code with a consistent interface for doing so. The actual
        mechanism will depend on which backend (Qt, Jupyter) is being used.

        The returned URL may be incomplete -- in Jupyter, it is not possible for
        the kernel to know the ultimate URL at which its server is accessible.
        Frontend code has to join these URLs with the server origin in order to
        obtain an actually retrievable URL.

        Data publishing may not be available -- in Jupyter, it requires a server
        extension, which may not be installed.

        The *extension* parameter is basically a hack to deal with some pywwt
        temporary files that are created on-disk without the "right" extension.
        """

        raise DataPublishingNotAvailableError()

    def _create_image_layer(self, **kwargs):
        """This method can be overridden to return specialized subclasses of
        :class:`~pywwt.layers.ImageLayer`. In particular, the Jupyter version
        of the viewer extends ``ImageLayer`` to add methods that add
        interactive UI controls for the layer parameters.

        """
        from .layers import ImageLayer

        return ImageLayer(self, **kwargs)

    # Startup

    async def becomes_ready(self, timeout=30):
        """
        An async function that finishes when the WWT app becomes ready.

        Returns
        -------
        self

        Notes
        -----

        The web browser running WWT needs to initialize the application and
        download some initial data files, which can take an indeterminate amount
        of time. This async function will complete once the WWT app has
        initialized and is ready to respond to control messages from pywwt.
        After this has happened the first time, this function will complete
        instantly.
        """

        if self._readinessAchieved:
            return self

        if self._readyFuture is None:
            loop = asyncio.get_running_loop()
            self._readyFuture = fut = loop.create_future()

            def maybe_time_it_out():
                if not fut.done():
                    self._readyFuture = None
                    fut.set_exception(asyncio.TimeoutError())

            loop.call_later(timeout, maybe_time_it_out)

        await self._readyFuture
        return self

    # Main attributes

    current_mode = None
    "The current rendering mode of the engine"

    imagery = None
    "Access to the engine's available imagesets"

    def _serve_tree(self, path):
        raise DataPublishingNotAvailableError()

    @property
    def instruments(self):
        """
        A list of instruments available for use in `add_fov`.
        """
        return self._instruments

    layers = None
    "Access to the active rendering layers"

    solar_system = None
    "Access to solar-system settings and data"

    # Settings.
    #
    # TODO: implement the remaining ones that haven't been wired up yet.

    actual_planet_scale = Bool(
        False,
        help="Whether to show planets to scale or as "
        "points with a fixed size "
        "(`bool`)",
    ).tag(wwt="actualPlanetScale", wwt_reset=True)

    alt_az_grid = Bool(
        False, help="Whether to show an altitude-azimuth grid " "(`bool`)"
    ).tag(wwt="showAltAzGrid", wwt_reset=True)

    # alt_az_text = Bool(False,
    #                    help='Whether to show labels for the altitude-azimuth grid\'s text '
    #                         '(`bool`)').tag(wwt='showAltAzGridText', wwt_reset=True)

    background = Unicode(
        "Hydrogen Alpha Full Sky Map",
        help="The layer to show in the background (`str`)",
    ).tag(wwt=None, wwt_reset=True)

    constellation_boundary_color = Color(
        "blue",
        help="The color of the constellation " "boundaries (`str` or " "`tuple`)",
    ).tag(wwt="constellationBoundryColor", wwt_reset=True)

    constellation_figure_color = Color(
        "red", help="The color of the constellation " "figure (`str` or " "`tuple`)"
    ).tag(wwt="constellationFigureColor", wwt_reset=True)

    constellation_selection_color = Color(
        "yellow",
        help="The color of the constellation " "selection (`str` or " "`tuple`)",
    ).tag(wwt="constellationSelectionColor", wwt_reset=True)

    constellation_boundaries = Bool(
        False,
        help="Whether to show boundaries for the "
        "selected constellations "
        "(`bool`)",
    ).tag(wwt="showConstellationBoundries", wwt_reset=True)

    constellation_figures = Bool(
        False, help="Whether to show the constellations " "(`bool`)"
    ).tag(wwt="showConstellationFigures", wwt_reset=True)

    constellation_selection = Bool(
        False,
        help="Whether to only show boundaries for "
        "the selected constellation "
        "(`bool`)",
    ).tag(wwt="showConstellationSelection", wwt_reset=True)

    # constellation_pictures = Bool(False,
    #                               help='Whether to show pictures of the constellations\' '
    #                                    'mythological representations '
    #                                    '(`bool`)').tag(wwt='showConstellationPictures', wwt_reset=True)

    # constellation_labels = Bool(False,
    #                             help='Whether to show labelss for constellations '
    #                                  '(`bool`)').tag(wwt='showConstellationLabels', wwt_reset=True)

    crosshairs = Bool(
        False, help="Whether to show crosshairs at the center of " "the field (`bool`)"
    ).tag(wwt="showCrosshairs", wwt_reset=True)

    crosshairs_color = Color(
        "white", help="The color of the crosshairs " "(`str` or `tuple`)"
    ).tag(wwt="crosshairsColor", wwt_reset=True)

    ecliptic = Bool(
        False, help="Whether to show the path of the ecliptic " "(`bool`)"
    ).tag(wwt="showEcliptic", wwt_reset=True)

    ecliptic_grid = Bool(
        False, help="Whether to show a grid relative to the " "ecliptic plane (`bool`)"
    ).tag(wwt="showEclipticGrid", wwt_reset=True)

    foreground = Unicode(
        "Digitized Sky Survey (Color)",
        help="The layer to show in the foreground (`str`)",
    ).tag(wwt=None, wwt_reset=True)

    foreground_opacity = Float(
        0.8, help="The opacity of the foreground layer " "(`float`)"
    ).tag(wwt=None, wwt_reset=True)

    galactic_mode = Bool(
        False,
        help="Whether the galactic plane should be horizontal "
        "in the viewer (`bool`)",
    ).tag(wwt="galacticMode", wwt_reset=True)

    galactic_grid = Bool(
        False, help="Whether to show a grid relative to the " "galactic plane (`bool`)"
    ).tag(wwt="showGalacticGrid", wwt_reset=True)

    # galactic_text = Bool(False,
    #                      help='Whether to show labels for the galactic grid\'s text '
    #                           '(`bool`)').tag(wwt='showGalacticGridText', wwt_reset=True)

    grid = Bool(False, help="Whether to show the equatorial grid " "(`bool`)").tag(
        wwt="showGrid", wwt_reset=True
    )

    local_horizon_mode = Bool(
        False,
        help="Whether the view should be that of "
        "a local latitude, longitude, and "
        "altitude (`bool`)",
    ).tag(wwt="localHorizonMode", wwt_reset=True)

    location_altitude = AstropyQuantity(
        0 * u.m,
        help="The altitude of the viewing "
        "location in local horizon mode "
        "(:class:`~astropy.units.Quantity`)",
    ).tag(wwt="locationAltitude", wwt_reset=True)

    location_latitude = AstropyQuantity(
        47.633 * u.deg,
        help="The latitude of the viewing "
        "location in local horizon mode "
        "(:class:`~astropy.units.Quantity`)",
    ).tag(wwt="locationLat", wwt_reset=True)

    location_longitude = AstropyQuantity(
        122.133333 * u.deg,
        help="The longitude of the viewing "
        "location in local horizon mode "
        "(:class:`~astropy.units.Quantity`)",
    ).tag(wwt="locationLng", wwt_reset=True)

    # Validators / observers for the settings above that need custom support.

    @observe("background")
    def _on_background_change(self, changed):
        self._send_msg(event="set_background_by_name", name=changed["new"])
        # Changing a layer resets the opacity, so we re-trigger the opacity setting
        self._send_msg(
            event="set_foreground_opacity", value=self.foreground_opacity * 100
        )

    @validate("background")
    def _validate_background(self, proposal):
        if proposal["value"] in self.available_layers:
            return proposal["value"]
        else:
            raise TraitError("background is not one of the available layers")

    @observe("foreground")
    def _on_foreground_change(self, changed):
        self._send_msg(event="set_foreground_by_name", name=changed["new"])
        # Changing a layer resets the opacity, so we re-trigger the opacity setting
        self._send_msg(
            event="set_foreground_opacity", value=self.foreground_opacity * 100
        )

    @validate("foreground")
    def _validate_foreground(self, proposal):
        if proposal["value"] in self.available_layers:
            return proposal["value"]
        else:
            raise TraitError("foreground is not one of the available layers")

    @observe("foreground_opacity")
    def _on_foreground_opacity_change(self, changed):
        self._send_msg(event="set_foreground_opacity", value=changed["new"] * 100)

    @validate("foreground_opacity")
    def _validate_foreground_opacity(self, proposal):
        if 0 <= proposal["value"] <= 1:
            return proposal["value"]
        else:
            raise TraitError("foreground_opacity should be between 0 and 1")

    @validate("location_altitude")
    def _validate_altitude(self, proposal):
        if proposal["value"].unit.physical_type == "length":
            return proposal["value"].to(u.meter)
        else:
            raise TraitError("location_altitude not in units of length")

    @validate("location_latitude")
    def _validate_latitude(self, proposal):
        if proposal["value"].unit.physical_type == "angle":
            return proposal["value"].to(u.degree)
        else:
            raise TraitError("location_latitude not in angle units")

    @validate("location_longitude")
    def _validate_longitude(self, proposal):
        if proposal["value"].unit.physical_type == "angle":
            return proposal["value"].to(u.degree)
        else:
            raise TraitError("location_longitude not in angle units")

    # Basic view controls

    def get_center(self):
        """
        Return the view's current right ascension and declination in degrees.
        """
        return SkyCoord(
            self._get_view_data("ra"),
            self._get_view_data("dec"),
            unit=(u.hourangle, u.deg),
        )

    def get_fov(self):
        """
        Return the view's current field of view in degrees.
        """
        return self._get_view_data("fov") * u.deg

    def center_on_coordinates(self, coord, fov=60 * u.deg, instant=True):
        """
        Center the view on a particular object or point in the sky.

        Parameters
        ----------
        coord : `~astropy.units.Quantity`
            The set of coordinates the view should center on.

        fov : `~astropy.units.Quantity`, optional
            The desired field of view.

        instant : `bool`, optional
            Whether the view changes instantly or smoothly scrolls to the
            desired location.
        """
        coord_icrs = coord.icrs
        self._send_msg(
            event="center_on_coordinates",
            ra=coord_icrs.ra.deg,
            dec=coord_icrs.dec.deg,
            fov=fov.to(u.deg).value,
            instant=instant,
        )

    def set_view(self, mode):
        """
        Change the view mode.

        Valid options include the default sky mode, a 3D universe mode with
        different viewing levels (the solar system, the Milky Way, and the
        observed universe), individual views of major solar system objects, and
        panoramas from lunar missions and NASA's Mars rovers.

        To find the list of available views, use the
        :attr:`~pywwt.BaseWWTWidget.available_views`.

        Parameters
        ----------
        mode : `str`
            The desired view mode. (default: 'sky')
        """

        mode = mode.lower()

        solar_system_mode = "3D Solar System View"

        if mode in VIEW_MODES_2D:
            if mode == "earth":
                # Switch to a daytime view of the earth
                mode = "Bing Maps Aerial"
            elif mode == "mars":
                mode = "Visible Imagery"
            self._send_msg(event="set_viewer_mode", mode=mode)
            self._last_sent_view_mode = mode
            if mode == "sky" or mode == "panorama":
                self.current_mode = mode
            else:
                self.current_mode = "planet"
        elif mode in VIEW_MODES_3D:
            self._send_msg(event="set_viewer_mode", mode=solar_system_mode)
            self.current_mode = mode
            self._last_sent_view_mode = solar_system_mode
        else:
            raise ValueError(
                "mode should be one of {0}".format(
                    "/".join(VIEW_MODES_2D + VIEW_MODES_3D)
                )
            )

        self.reset_view()

    def reset_view(self):
        """
        Reset the current view mode's coordinates and field of view to
        their original states.
        """
        if self.current_mode == "sky":
            self.center_on_coordinates(
                SkyCoord(0.0, 0.0, unit=u.deg), fov=60 * u.deg, instant=False
            )
        if self.current_mode == "planet":
            self.center_on_coordinates(
                SkyCoord(35.55, 11.43, unit=u.deg), fov=40 * u.deg, instant=False
            )
        if self.current_mode == "solar system":
            self.center_on_coordinates(
                SkyCoord(0.0, 0.0, unit=u.deg), fov=50 * u.deg, instant=False
            )
        if self.current_mode == "milky way":
            self.center_on_coordinates(
                SkyCoord(114.85, -29.52, unit=u.deg), fov=6e9 * u.deg, instant=False
            )
        if self.current_mode == "universe":
            self.center_on_coordinates(
                SkyCoord(16.67, 37.72, unit=u.deg), fov=1e14 * u.deg, instant=False
            )
        if self.current_mode == "panorama":
            pass

    @property
    def available_views(self):
        """
        A list of the modes that are currently available in the viewer.
        """
        return sorted(VIEW_MODES_2D + VIEW_MODES_3D)

    def reset(self):
        """
        Reset WWT to initial state.
        """

        # Remove any existing layers (not using a for loop since we're removing elements)
        while len(self.layers) > 0:
            self.layers[0].remove()

        # Reset coordinates to initial view
        gc = SkyCoord(0, 0, unit=("deg", "deg"), frame="icrs")
        self.center_on_coordinates(gc, 60 * u.deg)

        # Reset only traits with the wwt_reset tag
        for trait_name, trait in self.traits().items():
            if trait.metadata.get("wwt_reset"):
                setattr(self, trait_name, trait.default_value)

    # Clock controls

    def pause_time(self):
        """
        Pause the progression of time in the viewer.
        """
        self._send_msg(event="pause_time")

    def play_time(self, rate=1):
        """
        Resume the progression of time in the viewer.

        Parameters
        ----------
        rate : int or float
            The rate at which time passes (1 meaning real-time)
        """
        self._send_msg(event="resume_time", rate=rate)

    def get_current_time(self):
        """
        Return the viewer's current time as an `~astropy.time.Time` object.
        """
        return Time(self._get_view_data("datetime"), format="isot")

    def set_current_time(self, dt=None):
        """
        Set WWT's internal clock.

        Parameters
        ----------
        dt : `~datetime.datetime` or `~astropy.time.Time`
            A time, either as a `datetime.datetime` object or an
            astropy :class:`astropy.time.Time` object. If not specified, this
            uses the current time.

        Notes
        -----
        If you call this function and then immediately call
        :meth:`get_current_time`, the results will not necessarily agree. This
        is because this function has to send a command to WWT to tell it to
        update its internal clock, and in some environments this operation is
        not instantaneous.
        """
        # Ensure the object received is a datetime or Time; convert it to UTC
        utc_tm = ensure_utc(dt, str_allowed=False)
        self._send_msg(event="set_datetime", isot=utc_tm)

    # Data loading

    def load_image_collection(self, url, recursive=False, remote_only=False):
        """
        Load a collection of layers for possible use in the viewer.

        Parameters
        ----------
        url : `str`
            The URL of the desired image collection.

        recursive : optional `bool`
            If true, will also load any child folders referenced in the
            specified WTML file. The default behavior is only to load the
            single file and *not* recurse into subfolders. Recursive loading
            can potentially be very time-consuming.

        remote_only : optional `bool`
            ADASS hack! If true, will only send the load request to the app,
            and not attempt to open the URL locally.


        Notes
        -----
        The request to load the image collection must be relayed to the WWT
        JavaScript code, which will then issue a web request and process the
        response that it gets. Therefore, you can't rely on this function to
        take immediate effect; to use an image in a collection that you've
        loaded, you'll need to pause and give WWT time to receive and process
        your request.
        """
        # TODO: this isn't the right approach. We should get knowledge about
        # available layers from the app, since it might be instructed to load up
        # WTMLs by other clients that we don't even know about. Also, this
        # doesn't work for URLs that we serve ourselves, which may only be
        # fragments when we're running in the Jupyter context.
        if not remote_only:
            self._available_layers.update(get_imagery_layers(url))

        nest_asyncio.apply()
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._send_into_future(
            event="load_image_collection",
            url=url,
            loadChildFolders=recursive
        ))

    @property
    def available_layers(self):
        """
        A list of the layers that are currently available in the viewer.
        """
        return sorted(self._available_layers)

    # HiPS catalog support

    _available_hips_catalog_names = None

    @property
    def available_hips_catalog_names(self):
        """
        A list of the names of HiPS catalog layers that are currently available
        in the viewer.
        """
        return sorted(self._available_hips_catalog_names)

    # Support for source and HiPS catalog selection

    _most_recent_source = None

    @property
    def most_recent_source(self):
        """
        The most recent source selected in the viewer, represented as a dictionary.
        The items of this dictionary match the entries of the Source object detailed
        `here <https://docs.worldwidetelescope.org/webgl-reference/latest/apiref/research-app-messages/interfaces/selections.source.html>`_.
        """
        return self._most_recent_source

    _selected_sources = []

    @property
    def selected_sources(self):
        """
        A list of the selected sources, with each source represented as a dictionary.
        The items of these dictionaries match the entries of the Source object detailed
        `here <https://docs.worldwidetelescope.org/webgl-reference/latest/apiref/research-app-messages/interfaces/selections.source.html>`_.
        """
        return self._selected_sources

    # Annotations

    def clear_annotations(self):
        """
        Clears all annotations from the current view.
        """
        self._annotation_set.clear()
        return self._send_msg(event="clear_annotations")

    def add_circle(self, center=None, **kwargs):
        """
        Add a circle annotation to the current view.

        Parameters
        ----------
        center : `~astropy.units.Quantity`, optional
            The coordinates of desired center of the circle. If blank,
            defaults to the center of the current view.
        kwargs
            Optional arguments that allow corresponding Circle or Annotation
            attributes to be set upon shape initialization.
        """
        # TODO: could buffer JS call here
        circle = Circle(parent=self, center=center, **kwargs)
        return circle

    def add_polygon(self, points=None, **kwargs):
        """
        Add a polygon annotation to the current view.

        Parameters
        ----------
        points : `~astropy.units.Quantity`, optional
            The desired points that make up the polygon. If blank or just
            one point, the annotation will be initialized but will not be
            visible until more points are added. Note that the points should
            be specified in counter-clockwise order on the sky if you intend
            to fill the polygon.
        kwargs
            Optional arguments that allow corresponding Polygon or
            Annotation attributes to be set upon shape initialization.
        """
        # same TODO as above
        polygon = Polygon(parent=self, **kwargs)
        if points:
            polygon.add_point(points)
        return polygon

    def add_line(self, points=None, **kwargs):
        """
        Add a line annotation to the current view.

        Parameters
        ----------
        points : `~astropy.units.Quantity`, optional
            The desired points that make up the line. If blank or just one
            point, the annotation will be initialized but will not be
            visible until more points are added.
        kwargs
            Optional arguments that allow corresponding Line or Annotation
            attributes to be set upon shape initialization.
        """
        # same TODO as above
        line = Line(parent=self, **kwargs)
        if points:
            line.add_point(points)
        return line

    def add_collection(self, points, **kwargs):
        """
        Add a CircleCollection to the current view.

        Parameters
        ----------
        points : `~astropy.units.Quantity`
            The desired points that will serve as the centers of the
            circles that make up the collection. Requires at least two
            sets of coordinates for initialization.
        kwargs
            Optional arguments that allow corresponding Circle or
            Annotation attributes to be set upon shape initialization.
        """
        collection = CircleCollection(self, points, **kwargs)
        return collection

    # Tours

    def load_tour(self, url):
        """
        Load and begin playing a tour based on the URL to a .wtt file from
        the WorldWideTelescope website.

        Parameters
        ----------
        url : `str`
            The URL of the chosen tour -- must be a .wtt file.
        """
        # throw error if url doesn't end in .wtt
        if url[-4:] == ".wtt":
            self._send_msg(event="load_tour", url=url)
        else:
            raise ValueError("url must end in '.wwt'")

    def pause_tour(self):
        """
        Pause a loaded tour.
        """
        self._send_msg(event="pause_tour")

    def resume_tour(self):
        """
        Resume a paused tour.
        """
        self._send_msg(event="resume_tour")

    # Instrumental FOV support (built on the annotation support)

    def add_fov(self, telescope, center=None, rotate=0 * u.rad, **kwargs):
        """
        Add a telescope's field of view (FOV) to the current view.

        Parameters
        ----------
        telescope : `str`
            The telescope whose field of view will be displayed. Be sure to use
            the ``instruments`` attribute to see and select from the preset list
            of instruments available in pyWWT.
        center : `~astropy.units.Quantity`, optional
            The coordinates of desired center of the FOV. If blank,
            defaults to the center of the current view.
        rotate : `~astropy.units.Quantity`, optional
            The amount to rotate the FOV. Both radians and degrees are
            accepted. If blank, defaults to 0 radians (no rotation).
        kwargs
            Optional arguments that allow corresponding Polygon or
            Annotation attributes to be set upon shape initialization.
        """
        return FieldOfView(self, telescope, center, rotate, **kwargs)

    # HTML (interactive figure) export

    def save_as_html_bundle(self, dest, title=None, max_width=None, max_height=None):
        """
        Save the current view as a web page with supporting files.

        This feature is currently under development, so not all
        settings/features that can be set in pyWWT will be saved

        Parameters
        ----------
        dest : `str`
            The path to output the bundle to. The path must represent a
            directory (which will be created if it does not exist) or a zip file.
        title : `str`, optional
            The desired title for the HTML page. If blank, a generic title will be used.
        max_width : `int`, optional
            The maximum width of the WWT viewport on the exported HTML page in pixels.
            If left blank, the WWT viewport will fill the enitre width of the browser.
        max_height : `int`, optional
            The maximum height of the WWT viewport on the exported HTML page in pixels.
            If left blank, the WWT viewport will fill the enitre height of the browser.
        """
        dest_root, dest_extension = os.path.splitext(dest)
        if dest_extension and dest_extension != ".zip":
            raise ValueError("'dest' must be either a directory or a .zip file")

        is_compressed = dest_extension == ".zip"
        if is_compressed:
            figure_dir = tempfile.mkdtemp()
        else:
            if not os.path.exists(dest):
                os.makedirs(os.path.abspath(dest))
            figure_dir = dest

        fig_src_dir = os.path.join(os.path.dirname(__file__), "interactive_figure")
        shutil.copy(os.path.join(fig_src_dir, "index.html"), figure_dir)
        script_dir = os.path.join(figure_dir, "scripts")
        if not os.path.exists(script_dir):
            os.mkdir(script_dir)
        shutil.copy(os.path.join(fig_src_dir, "interactive_figure.js"), script_dir)
        shutil.copy(os.path.join(fig_src_dir, "wwt_json_api.js"), script_dir)

        self._serialize_to_json(
            os.path.join(figure_dir, "wwt_figure.json"), title, max_width, max_height
        )

        if len(self.layers) > 0:
            data_dir = os.path.join(figure_dir, "data")
            if not os.path.exists(data_dir):
                os.mkdir(data_dir)
            self._save_added_data(data_dir)

        if is_compressed:
            zip_parent_dir = os.path.abspath(os.path.dirname(dest_root))
            if not os.path.exists(zip_parent_dir):
                os.makedirs(zip_parent_dir)
            shutil.make_archive(dest_root, "zip", root_dir=figure_dir)

    def _serialize_state(self, title, max_width, max_height):
        state = dict()
        state["html_settings"] = {
            "title": title,
            "max_width": max_width,
            "max_height": max_height,
        }

        state["wwt_settings"] = {}
        for trait in self.traits().values():
            wwt_name = trait.metadata.get("wwt")
            if wwt_name:
                trait_val = trait.get(self)
                if isinstance(trait_val, u.Quantity):
                    trait_val = trait_val.value
                state["wwt_settings"][wwt_name] = trait_val

        center = self.get_center()
        fov = self.get_fov()
        state["view_settings"] = {
            "mode": self._last_sent_view_mode,
            "ra": center.icrs.ra.deg,
            "dec": center.icrs.dec.deg,
            "fov": fov.to_value(u.deg),
        }

        state["foreground_settings"] = {
            "foreground": self.foreground,
            "background": self.background,
            "foreground_alpha": self.foreground_opacity * 100,
        }

        state["layers"] = self.layers._serialize_state()

        if self.current_mode in VIEW_MODES_3D:
            self.solar_system._add_settings_to_serialization(state)

        state["annotations"] = []
        for annot in self._annotation_set:
            state["annotations"].append(annot._serialize_state())
        return state

    def _serialize_to_json(self, file, title, max_width, max_height):
        state = self._serialize_state(title, max_width, max_height)
        with open(file, "w") as file_obj:
            json.dump(state, file_obj)

    def _save_added_data(self, dir):
        self.layers._save_all_data_for_serialization(dir)



"""
Routines for the pywwt notebook server extension.

Most users of pywwt can ignore this module.

In order to make files available to the WWT engine, we need to serve them over
HTTP. Most of this is now done using the ``wwt_kernel_data_relay`` server
extension, but we also provide a separate server extension that provides a
static copy of the WWT research app, for environments where security cookies
prevent an externally-hosted version of the app (i.e., wwtassets.org) from
accessing data served up by the kernel.
"""

import os
import mimetypes
from tornado import web
from notebook.utils import url_path_join
from notebook.base.handlers import IPythonHandler

__all__ = [
    "load_jupyter_server_extension",
]


STATIC_DIR = os.path.join(os.path.dirname(__file__), "web_static")


class WWTStaticFileHandler(IPythonHandler):
    def get(self, filename):
        static_path = os.path.join(STATIC_DIR, filename)

        if os.path.isdir(static_path):
            path = os.path.join(static_path, "index.html")
            filename = "index.html"  # for mime-type guess below
        else:
            path = static_path

        # Do our best to set an appropriate Content-Type.
        content_type = mimetypes.guess_type(filename)[0]
        if content_type is None:
            content_type = "application/binary"
        self.set_header("Content-Type", content_type)

        # Add wide-open CORS headers to allow external WWT apps to access data.
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Methods", "GET,HEAD")
        self.set_header(
            "Access-Control-Allow-Headers",
            "Content-Disposition,Content-Encoding,Content-Length,Content-Type",
        )

        try:
            with open(path, "rb") as f:
                content = f.read()
        except FileNotFoundError:
            raise web.HTTPError(404)

        self.finish(content)


def load_jupyter_server_extension(nb_server_app):
    """
    A support function used to integrate pywwt with Jupyter.
    """
    web_app = nb_server_app.web_app
    host_pattern = ".*$"

    mimetypes.add_type("image/fits", ".fits")
    mimetypes.add_type("image/fits", ".fts")
    mimetypes.add_type("image/fits", ".fit")

    route_pattern = url_path_join(web_app.settings["base_url"], "/wwtstatic/(.*)")
    web_app.add_handlers(host_pattern, [(route_pattern, WWTStaticFileHandler)])



#! /usr/bin/env python
# -*- mode: python; coding: utf-8 -*-
# Copyright (c) Jupyter Development Team, AAS WorldWide Telescope team
# Distributed under the terms of the Modified BSD License.

from __future__ import print_function
import io
from os.path import join as pjoin
import sys

sys.path.append(".")
from setupbase import (
    create_cmdclass,
    install_npm,
    ensure_targets,
    find_packages,
    combine_commands,
    get_version,
    HERE,
)

from setuptools import setup

# Gather package metadata.

name = "pywwt"
version = get_version(pjoin(name, "_version.py"))

with io.open("README.md", encoding="utf_8") as f:
    LONG_DESCRIPTION = f.read()

# Wire up the NPM build to the Python package build -- we generate JavaScript
# artifacts from the `frontend/` package and then include them as data assets in
# the pywwt Python package.

nb_path = pjoin(name, "nbextension", "static")

js_outputs = [
    pjoin(nb_path, "index.js"),
]

js_content_command = combine_commands(
    install_npm("frontend", build_cmd="pywwt-export"),
    ensure_targets(js_outputs),
)

# Custom "command class" that (1) makes sure to create the JS content, (2)
# includes that content as extra "package data" in the Python package, and (3)
# can install special metadata files in the Python environment root.

package_data_spec = {
    name: [
        "interactive_figure/*.html",
        "interactive_figure/*.js",
        "labextension/*.tgz",
        "nbextension/static/*.*js*",
        "tests/data/*/*.png",
        "tests/data/*.fits",
        "web_static/**",
    ]
}

lab_path = pjoin(name, "labextension")
local_nbdotd_path = pjoin("jupyter.d", "notebook.d")
local_jnbcfgdotd_path = pjoin("jupyter.d", "jupyter_notebook_config.d")

data_files_spec = [
    ("share/jupyter/nbextensions/pywwt", nb_path, "*.js*"),
    ("share/jupyter/lab/extensions", lab_path, "*.tgz"),
    ("etc/jupyter/nbconfig/notebook.d", local_nbdotd_path, "pywwt.json"),
    ("etc/jupyter/jupyter_notebook_config.d", local_jnbcfgdotd_path, "pywwt.json"),
]

cmdclass = create_cmdclass(
    "js-content", package_data_spec=package_data_spec, data_files_spec=data_files_spec
)
cmdclass["js-content"] = js_content_command

# Ready to go!

setup_args = dict(
    name=name,
    description="The AAS WorldWide Telescope from Python",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    version=version,
    cmdclass=cmdclass,
    packages=find_packages(),
    author="AAS WorldWide Telescope  team",
    author_email="wwt@aas.org",
    url="https://github.com/WorldWideTelescope/pywwt",
    license="BSD",
    platforms="Linux, Mac OS X, Windows",
    keywords=["Jupyter", "Widgets", "IPython"],
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Topic :: Multimedia :: Graphics",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Framework :: Jupyter",
    ],
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=[
        # Keep alphabetized:
        "astropy>=1.0,!=4.0.1",
        "beautifulsoup4",
        "ipyevents",
        "ipywidgets>=7.0.0",
        "lxml",
        "matplotlib>1.5",
        "nest_asyncio",
        "numpy>=1.9",
        "python-dateutil",
        "pytz",
        "qtpy",
        "reproject>=0.8",
        "requests",
        "toasty>=0.15",
        "tornado",
        "traitlets>=5",
        "wwt_data_formats>=0.12",
    ],
    extras_require={
        "test": [
            "pytest",
            "pytest-cov>=2.6.1",
            "pytest-remotedata>=0.3.1",
        ],
        "docs": [
            "astropy-sphinx-theme",
            "sphinx>=1.6",
            "sphinx-automodapi",
            "numpydoc",
            "jupyter_sphinx",
        ],
        "qt": [
            'PyQt5;python_version>="3"',
            'PyQtWebEngine;python_version>="3"',
        ],
        "lab": ["jupyterlab"],
    },
    entry_points={},
)

if __name__ == "__main__":
    setup(**setup_args)



# Copyright 2021 the .NET Foundation
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
A simple helper to make it possible to run `python -m pywwt.tests` to analyze
the image-test corpus -- see README.md.
"""

import sys
from . import analyze_test_image_corpus
analyze_test_image_corpus(sys.argv[1:])



version_info = (0, 15, 2, 'final', 0)  # cranko project-version tuple

_specifier_ = {'alpha': '.a', 'beta': '.b', 'candidate': '.rc', 'final': '', 'dev': '.dev'}

__version__ = '%s.%s.%s%s' % (
    version_info[0],
    version_info[1],
    version_info[2],
    '' if version_info[3] == 'final' else _specifier_[version_info[3]]+str(version_info[4])
)



# Copyright 2021 the .NET Foundation
# Licensed under the three-clause BSD License

"""
Utilities for dealing with WWT imagery collections.

The code in this module is fairly old and in need of updating. You won't need to
use its functionality directly if you're not a pywwt developer.
"""

import re

from io import BytesIO
from collections import OrderedDict
from xml.etree.ElementTree import ElementTree

import requests

__all__ = [
    "get_imagery_layers",
    "Bandpass",
    "ImageryLayers",
]


def get_imagery_layers(url):
    """
    Get the list of available image layers that can be used as background
    or foreground based on the URL to a WTML (WorldWide Telescope image
    collection file).

    Parameters
    ----------
    url : `str`
        The URL of the image collection.
    """

    available_layers = OrderedDict()

    # Get the XML describing the available surveys
    response = requests.get(url)
    assert response.ok
    b = BytesIO(response.content)
    e = ElementTree()
    t = e.parse(b)

    for survey in t.iter("ImageSet"):
        name = survey.attrib["Name"]
        thumbnail_url = survey.find("ThumbnailUrl")
        if thumbnail_url is not None:
            thumbnail_url = getattr(thumbnail_url, "text", None) or None
        available_layers[name] = {"thumbnail": thumbnail_url}

    return available_layers


class ImageryLayers:
    """
    A supplemental class that standardizes layer names, sorts them by
    bandpass, and makes it possible to tab complete them.
    """

    def __init__(self, layer_list):
        self._layers = {}
        self._spectrum = [
            "gamma",
            "x",
            "uv",
            "visible",
            "ir",
            "micro",
            "radio",
            "other",
        ]
        self.integers = [
            "zero",
            "one",
            "two",
            "three",
            "four",
            "five",
            "six",
            "seven",
            "eight",
            "nine",
        ]
        for band in self._spectrum:
            self._layers[band] = {}

        self._list2dict(layer_list)

    def _list2dict(self, og_list):
        # Helps turn the list of layer names used to initialize the class
        # (og_list) into a dict.
        for layer in og_list:

            if re.search(r"(?i)gamma", layer) is not None:
                self._add2dict(self._layers, layer, "gamma")
                continue  # automatically advance to next iteration

            if re.search(r"(?i)x(-|\s)?ray", layer) is not None:
                self._add2dict(self._layers, layer, "x")
                continue

            if (
                re.search(r"(?i)ultra(-|\s)?violet", layer) is not None
                or re.search(r"(?i)[^\d\w]+uv|uv[^\d\w]+", layer) is not None
            ):
                self._add2dict(self._layers, layer, "uv")
                continue

            if (
                re.search(r"(?i)optical", layer) is not None
                or re.search(r"(?i)visible", layer) is not None
            ):
                self._add2dict(self._layers, layer, "visible")
                continue

            if (
                re.search(r"(?i)infrared", layer) is not None
                or re.search(r"(?i)[^\d\w]+ir|ir[^\d\w]+", layer) is not None
            ):
                self._add2dict(self._layers, layer, "ir")
                continue

            if (
                re.search(r"(?i)microwave", layer) is not None
                or re.search(r"(?i)[^\d\w]+cmb|cmb[^\d\w]+", layer) is not None
            ):
                self._add2dict(self._layers, layer, "micro")
                continue

            if re.search(r"(?i)radio", layer) is not None:
                self._add2dict(self._layers, layer, "radio")
                continue

            self._add2dict(self._layers, layer, "other")

    def _add2dict(self, diction, full_layer, bandpass):
        # Handles a layer's (full_layer) actual addition to the master
        # dict of layers (diction) according to its bandpass.
        suffix = ""
        short = self._shorten(full_layer) + suffix

        while short in diction[bandpass]:
            if suffix:
                suffix += 1
                short = short[:-1] + str(suffix)
            else:
                suffix = 1
                short += str(suffix)

        diction[bandpass][short] = {}
        diction[bandpass][short]["full_name"] = full_layer

    def _shorten(self, string):
        # Unlocks tab completion by shortening a full layer's name
        # (string) to a valid Python name based on its first word.
        cut_left = re.search(r"^[_\W]+", string)
        if cut_left is not None:
            string = string[cut_left.end():]

        cut_right = re.search(r"[_\W]", string)
        if cut_right is not None:
            string = string[: cut_right.start()].lower()

        digit = re.search(r"^\d", string)
        if digit is not None:
            for i, num in enumerate(self.integers, 0):
                if str(i) == digit.group(0):
                    string = string[: digit.start()] + num + string[digit.end():]
                    break

        return string

    def __dir__(self):
        return sorted(self._layers.keys())

    def __getattr__(self, band):
        return Bandpass(self._layers[band])


class Bandpass:
    """
    Allows the __getattr__() method from ImageryLayers to reach into the
    inner dict corresponding to layers of a particular bandpass.
    """

    def __init__(self, band):
        self._band = band

    def __dir__(self):
        return sorted(self._band.keys())

    def __getattr__(self, name):
        return self._band[name]["full_name"]
